/***********************************************************
* API function file for Phoenix Lexical Analyzer Generator *
************************************************************

A regular expression is used to define each lexical unit, symbolized by a token.

Syntax rules for a regular expression:
-------------------------------------

1. A character is normally represented literally. The special character \ allows for escaped forms, as in C:
   \t,\n,\r,\f,\0,\ddd (octal),\xdd(hex).
2. The operator for concatenation is implicit: operands just stand next to one another.
3. The operator for alternatives ("or"), is represented by |.
4. A compact representation of a sequence like a|b|c|d, with many alternatives, is the class representation:
   [a-d], where [ and ] serve as delimiters. Inside a class, the special character - indicates a range of
   characters, bounds included. The character \ allows to escape the following character, as in C (see 1.).
   Inside a class, it is of particular interest for \\, \- and \]. The lower bound must figure first.
   A class may contain any mix of characters and character ranges, as in: [A-Za-z0-9_\]], which means:
   any uppercase or lowercase letter, or digit, or the underscore, or the right square bracket.
   A minus sign figuring as the first or last character of a class does not represent a range, rather the
   minus sign itself, as in [+-].
5. A string is a sequence of characters enclosed in double quotes. It represents the concatenation of these
   characters, each taken literally. The only characters you may escape in a string are \ and ", as in:
   "abc\"", or ",;\\".
6. The operator ?, placed at the right of it's unique operand, means: zero or one instance of the preceding
   expression.
7. The operator *, placed at the right of it's unique operand, means: zero or n instances of the preceding
   expression.
8. The operator +, placed at the right of it's unique operand, means: one or n instances of the preceding
   expression.
9. A . (period), when not escaped, stands for "any character but newline", newline being ANSI 10.
10. The operator precedence rules are as follows, from highest to lowest:
        ?, *, +
        concatenation (implicit operator)
        |
    To change the precedence, use the parenthesis ( ). Note that only concatenation and | take two operands,
    and concatenation distributes over |. Example: x(abc|def) is equivalent to xabc|xdef.
11. Beware of spaces and tabulations: they have a meaning, so avoid using them in an expression. However,
    to represent a space, use [ ], a class limited to a space. Use \t for a tabulation.
********************************************************************************************************************/

#include <windows.h>
#include "ph_scan.h"        // Lexical analyzer generator
#include "ph_parse.h"       // Parser generator
#include "ph_error.h"

#define  APPNAME            "PHOENIX"

HINSTANCE hDLLInst;                             // The handle of the DLL instance
static unsigned char FileName[MAX_FILE_NAME+1]; // Work variable for file names
#define SOURCE_LEXICAL      1
#define SOURCE_PARSER       2
WORD SourceType;                                // SOURCE_LEXICAL or SOURCE_PARSER
BOOL Packed;                                    // TRUE if DFA is packed

/* DLL Initialization */
int CALLBACK LibMain(HINSTANCE hinst, WORD wDataSeg, WORD cbHeapSize, LPSTR lpszCmdLine)
{
  hDLLInst = hinst;

  return 1;
}

#define WSOUT(hf,s)         if(WriteToFile(hf,(LPSTR)s)==FALSE)return FALSE
#define WSOUTCLOSE(hf,s)    if(WriteToFile(hf,(LPSTR)s)==FALSE){_lclose(hf);return FALSE;}

/* Copy a null terminated string */
void CopyHugeString(LPSTR lpDest, LPHUGE hpSource)
{
	while(*hpSource != '\0')
	{
		*lpDest++ = *hpSource++;
	}
	*lpDest = '\0';
}

/* Copy a non-null terminated string. No null terminator is appended. */
void CopyHugeStringN(LPSTR lpDest, LPHUGE hpSource, unsigned int uiLength)
{
	register unsigned int i;
	for(i = 0; i < uiLength; i++)
	{
		*lpDest++ = *hpSource++;
	}
}

/* Writing a string to a file */
BOOL WriteToFile(HFILE hf,LPSTR Buf)
{
  if(_lwrite(hf,(LPSTR)Buf,lstrlen(Buf))== HFILE_ERROR)
  {
    LexErrorCode = LEXERR_CANT_WRITE_FILE;
    return FALSE;
  }

  return TRUE;
}

BOOL PhoenixOutputLexTokens(HFILE hf)
{
  register WORD i;
  UINT Resource;
  int TokenValue;
  unsigned char memo;

  /*** Token table ***/

  switch(Target)
  {
    case TARGET_WIN31:
        Resource = WIN31_OUT_RESOURCE_TOKEN;
        break;
    case TARGET_UNIX:
        Resource = UNIX_OUT_RESOURCE_TOKEN;
        break;
  }

  LoadString(hDLLInst,Resource,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUT(hf,OutBuffer);

  for(i = 0; i <= TotalStates; i++)
  {
    if(i % 10 == 0) WSOUT(hf,"\r\n");

    TokenValue = hpGNextState[i].Token;
    if(TokenValue && hpP2RegExp[TokenValue-1].Action.CodeSize < 0)
    {
        /* A simple return token action */
        memo = hpP2RegExp[TokenValue-1].Action.hpCodeStart[-hpP2RegExp[TokenValue-1].Action.CodeSize];
        hpP2RegExp[TokenValue-1].Action.hpCodeStart[-hpP2RegExp[TokenValue-1].Action.CodeSize] = '\0';

    	CopyHugeString((LPSTR)OutBuffer, hpP2RegExp[TokenValue-1].Action.hpCodeStart);
        if(i != TotalStates)
        	lstrcat((LPSTR)OutBuffer, ",");

        hpP2RegExp[TokenValue-1].Action.hpCodeStart[-hpP2RegExp[TokenValue-1].Action.CodeSize] = memo;
    }
    else
    {
        if(TokenValue == 0)
        {
            /* A user action */
            if(i != TotalStates)
                wsprintf((LPSTR)OutBuffer,"0,");
            else
                wsprintf((LPSTR)OutBuffer,"0");
        }
        else
        {
            /* A user action */
            if(i != TotalStates)
                wsprintf((LPSTR)OutBuffer,"-%u,",TokenValue);
            else
                wsprintf((LPSTR)OutBuffer,"-%u",TokenValue);
        }
    }

    WSOUT(hf,OutBuffer);
  }

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUT(hf,"\r\nEND\r\n");
        break;
    case TARGET_UNIX:
        WSOUT(hf,"\r\n};\r\n");
        break;
  }

  return TRUE;
}

BOOL PhoenixOutputLexTables(LPSTR Directory)
{
  WORD i,j;
  UINT Resource;
  HFILE hf;
  OFSTRUCT ofs;
  LPSTR Out;

  lstrcpy((LPSTR)FileName,Directory);
  switch(Target)
  {
    case TARGET_WIN31:
        lstrcat((LPSTR)FileName,"\\winscan.rc");                // Lex tables resource file
        break;
    case TARGET_UNIX:
        lstrcat((LPSTR)FileName,"\\winscant.h");                // Lex tables header file
        break;
  }

  for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i]; // Copy FileName to parameter
  LexErrorStringParam[i] = 0;   // Truncate to MAX_ERROR_PARAM

  hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
  if (hf == HFILE_ERROR)
  {
    LexErrorCode = LEXERR_CANT_CREATE_FILE;
    return FALSE;
  }

  switch(Target)
  {
    case TARGET_WIN31:
        LoadString(hDLLInst,WIN31_OUT_RESOURCE_TITLE,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
        WSOUTCLOSE(hf,OutBuffer);
        break;
    case TARGET_UNIX:
        LoadString(hDLLInst,UNIX_OUT_RESOURCE_TITLE,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
        WSOUTCLOSE(hf,OutBuffer);
        break;
  }

  LoadString(hDLLInst,OUT_RESOURCE_COUNT,(LPSTR)OutParam,MAX_OUT_BUFFER);
  wsprintf((LPSTR)OutBuffer,(LPSTR)OutParam,TotalStates);
  WSOUT(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        Resource = WIN31_OUT_RESOURCE_NEXT_STATE;
        break;
    case TARGET_UNIX:
        Resource = UNIX_OUT_RESOURCE_NEXT_STATE;
        break;
  }

  LoadString(hDLLInst,Resource,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  for(i = 0; i <= TotalStates; i++)
  {
    Out = (LPSTR)OutBuffer;
    if(i != 0)
    {
        wsprintf(Out,"\r\n");
        Out += 2;
    }

    if(i == DfaStart)
        wsprintf(Out,"\r\n/* State %u (Start State) */",i);
    else
        wsprintf(Out,"\r\n/* State %u */",i);
    while(*Out) Out++;
    for(j = 0; j < MAX_CHARS; j++)
    {
        if(j % 50 == 0)
        {
            *Out++ = '\r'; *Out++ = '\n'; *Out = 0;

            WSOUTCLOSE(hf,OutBuffer);

            Out = (LPSTR)OutBuffer;
            *Out = 0;
        }
        if(i == TotalStates && j == MAX_CHARS - 1)
            wsprintf(Out,"%u",hpGNextState[i].NextState[j]);
        else
        {
            wsprintf(Out,"%u,",hpGNextState[i].NextState[j]);
            while(*Out) Out++;
        }
    }
    WSOUTCLOSE(hf,OutBuffer);
  }

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n#include \"wstoken.h\"\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n");
        break;
  }

  /* Token table */
  if(PhoenixOutputLexTokens(hf) == FALSE) {_lclose(hf); return FALSE;}

  _lclose(hf);

  return TRUE;
}

BOOL PhoenixOutputPackedTables(LPSTR Directory)
{
  WORD i;
  UINT Resource1,Resource2,Resource3,Resource4;
  HFILE hf;
  OFSTRUCT ofs;
  LPSTR Out;

  lstrcpy((LPSTR)FileName,Directory);
  switch(Target)
  {
    case TARGET_WIN31:
        lstrcat((LPSTR)FileName,"\\winscan.rc");                // Lex tables resource file
        break;
    case TARGET_UNIX:
        lstrcat((LPSTR)FileName,"\\winscant.h");                // Lex tables header file
        break;
  }

  for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i]; // Copy FileName to parameter
  LexErrorStringParam[i] = 0;   // Truncate to MAX_ERROR_PARAM

  hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
  if (hf == HFILE_ERROR)
  {
    LexErrorCode = LEXERR_CANT_CREATE_FILE;
    return FALSE;
  }

  switch(Target)
  {
    case TARGET_WIN31:
        LoadString(hDLLInst,WIN31_OUT_RESOURCE_TITLE,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
        WSOUTCLOSE(hf,OutBuffer);
        break;
    case TARGET_UNIX:
        LoadString(hDLLInst,UNIX_OUT_RESOURCE_TITLE,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
        WSOUTCLOSE(hf,OutBuffer);
        break;
  }

  LoadString(hDLLInst,OUT_RESOURCE_COUNT,(LPSTR)OutParam,MAX_OUT_BUFFER);
  wsprintf((LPSTR)OutBuffer,(LPSTR)OutParam,TotalStates);
  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        Resource1 = WIN31_OUT_RESOURCE_DEFAULT;
        Resource2 = WIN31_OUT_RESOURCE_BASE;
        Resource3 = WIN31_OUT_RESOURCE_NEXT;
        Resource4 = WIN31_OUT_RESOURCE_CHECK;
        break;
    case TARGET_UNIX:
        Resource1 = UNIX_OUT_RESOURCE_DEFAULT;
        Resource2 = UNIX_OUT_RESOURCE_BASE;
        Resource3 = UNIX_OUT_RESOURCE_NEXT;
        Resource4 = UNIX_OUT_RESOURCE_CHECK;
        break;
  }

  /*** Default array ***/

  LoadString(hDLLInst,Resource1,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  Out = (LPSTR)OutBuffer;
  for(i = 0; i <= TotalStates; i++)
  {
    if(i % 10 == 0)
    {
        wsprintf(Out,"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
        Out = (LPSTR)OutBuffer; *Out = 0;
    }
    if(i < TotalStates)
        wsprintf(Out,"%u,",PDefault[i]);
    else
        wsprintf(Out,"%u",PDefault[i]);
    while(*Out) Out++;
  }

  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n\r\n");
        break;
  }

  /*** Base array ***/

  LoadString(hDLLInst,Resource2,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  Out = (LPSTR)OutBuffer;
  for(i = 0; i <= TotalStates; i++)
  {
    if(i % 10 == 0)
    {
        wsprintf(Out,"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
        Out = (LPSTR)OutBuffer; *Out = 0;
    }
    if(i < TotalStates)
        wsprintf(Out,"%u,",PBase[i]);
    else
        wsprintf(Out,"%u",PBase[i]);
    while(*Out) Out++;
  }

  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n\r\n");
        break;
  }

  /*** Next array ***/

  wsprintf(OutBuffer,"/* The PNext and PCheck arrays have %u entries */",nCheck);
  WSOUTCLOSE(hf,OutBuffer);

  LoadString(hDLLInst,Resource3,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  Out = (LPSTR)OutBuffer;
  for(i = 0; i < nCheck; i++)
  {
    if(i % 10 == 0)
    {
        wsprintf(Out,"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
        Out = (LPSTR)OutBuffer; *Out = 0;
    }
    if(i < nCheck - 1)
        wsprintf(Out,"%u,",PNext[i]);
    else
        wsprintf(Out,"%u",PNext[i]);
    while(*Out) Out++;
  }

  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n\r\n");
        break;
  }

  /*** Check array ***/

  LoadString(hDLLInst,Resource4,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  Out = (LPSTR)OutBuffer;
  for(i = 0; i < nCheck; i++)
  {
    if(i % 10 == 0)
    {
        wsprintf(Out,"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
        Out = (LPSTR)OutBuffer; *Out = 0;
    }
    if(i < nCheck - 1)
        wsprintf(Out,"%u,",(PCheck[i] == VIRGIN ? 0 : PCheck[i]));
    else
        wsprintf(Out,"%u",(PCheck[i] == VIRGIN ? 0 : PCheck[i]));
    while(*Out) Out++;
  }

  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n#include \"wstoken.h\"\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n\r\n");
        break;
  }

  /* Token table */
  if(PhoenixOutputLexTokens(hf) == FALSE) {_lclose(hf); return FALSE;}

  _lclose(hf);

  return TRUE;
}

BOOL PhoenixOutputPackedActionGoto(LPSTR Directory)
{
  WORD i;
  UINT Resource1,Resource2,Resource3;
  HFILE hf;
  OFSTRUCT ofs;
  LPSTR Out;

  lstrcpy((LPSTR)FileName,Directory);
  switch(Target)
  {
    case TARGET_WIN31:
        lstrcat((LPSTR)FileName,"\\wsparse.rc");            // Parsing tables resource file
        break;
    case TARGET_UNIX:
        lstrcat((LPSTR)FileName,"\\wsparse.h");             // Parsing tables header file
        break;
  }

  for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i]; // Copy FileName to parameter
  LexErrorStringParam[i] = 0;   // Truncate to MAX_ERROR_PARAM

  hf = OpenFile(FileName, &ofs, OF_WRITE);  // Open the file for append
  if (hf == HFILE_ERROR)
  {
    LexErrorCode = LEXERR_CANT_CREATE_FILE;
    return FALSE;
  }

  /* Go to the end of the file */
  if(_llseek(hf,0L,2) == -1)
  {
    LexErrorCode = LEXERR_CANT_WRITE_FILE;
    return FALSE;
  }

  switch(Target)
  {
    case TARGET_WIN31:
        Resource1 = WIN31_OUT_RESOURCE_ACTIONSTATE;
        Resource2 = WIN31_OUT_RESOURCE_GOTOSYMBOL;
        Resource3 = WIN31_OUT_RESOURCE_AGPOOL;
        break;
    case TARGET_UNIX:
        Resource1 = UNIX_OUT_RESOURCE_ACTIONSTATE;
        Resource2 = UNIX_OUT_RESOURCE_GOTOSYMBOL;
        Resource3 = UNIX_OUT_RESOURCE_AGPOOL;
        break;
  }

  /*** ActionState array ***/

  wsprintf(OutBuffer,"\r\n/* ActionState has %u entries */",nCsi);
  WSOUTCLOSE(hf,OutBuffer);
  LoadString(hDLLInst,Resource1,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  Out = (LPSTR)OutBuffer;
  for(i = 0; i < nCsi; i++)
  {
    if(i % 10 == 0)
    {
        wsprintf(Out,"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
        Out = (LPSTR)OutBuffer; *Out = 0;
    }
    if(i < nCsi - 1)
        wsprintf(Out,"%d,",ActionState[i]);
    else
        wsprintf(Out,"%d",ActionState[i]);
    while(*Out) Out++;
  }

  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n\r\n");
        break;
  }

  /*** GotoSymbol array ***/

  wsprintf(OutBuffer,"/* GotoSymbol has %u entries */",nSymbol - nTerm);
  WSOUTCLOSE(hf,OutBuffer);
  LoadString(hDLLInst,Resource2,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  Out = (LPSTR)OutBuffer;
  for(i = nTerm; i < nSymbol; i++)
  {
    if((i - nTerm) % 10 == 0)
    {
        wsprintf(Out,"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
        Out = (LPSTR)OutBuffer; *Out = 0;
    }
    if(i < nSymbol - 1)
        wsprintf(Out,"%u,",GotoSymbol[i - nTerm]);
    else
        wsprintf(Out,"%u",GotoSymbol[i - nTerm]);
    while(*Out) Out++;
  }

  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n\r\n");
        break;
  }

  /*** AGPool array ***/

  wsprintf(OutBuffer,"/* AGPool has %u entries */",nAGPool * 2);
  WSOUTCLOSE(hf,OutBuffer);

  LoadString(hDLLInst,Resource3,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  Out = (LPSTR)OutBuffer;
  for(i = 0; i < nAGPool; i++)
  {
    if(i % 10 == 0)
    {
        wsprintf(Out,"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
        Out = (LPSTR)OutBuffer; *Out = 0;
    }

    if(i < nAGPool - 1)
        wsprintf(Out,"%d,%d,",(short int)LOWORD(AGPool[i]),(short int)HIWORD(AGPool[i]));   // Current state, then next state
    else
        wsprintf(Out,"%d,%d", (short int)LOWORD(AGPool[i]),(short int)HIWORD(AGPool[i]));   // Current state, then next state
    while(*Out) Out++;
  }

  WSOUTCLOSE(hf,OutBuffer);

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n\r\n");
        break;
  }

  _lclose(hf);

  return TRUE;
}

/* Generation of the Production arrays (production's length and left side nonterminal) */
BOOL PhoenixOutputProductions(LPSTR Directory)
{
  WORD i;
  UINT Resource;
  HFILE hf;
  OFSTRUCT ofs;

  lstrcpy((LPSTR)FileName,Directory);
  switch(Target)
  {
    case TARGET_WIN31:
        lstrcat((LPSTR)FileName,"\\wsparse.rc");            // Parser's production tables resource file
        break;
    case TARGET_UNIX:
        lstrcat((LPSTR)FileName,"\\wsparse.h");             // Parser's production tables header file
        break;
  }

  for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i]; // Copy FileName to parameter
  LexErrorStringParam[i] = 0;   // Truncate to MAX_ERROR_PARAM

  hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
  if (hf == HFILE_ERROR)
  {
    LexErrorCode = LEXERR_CANT_CREATE_FILE;
    return FALSE;
  }

  LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  /*** Left side nonterminal of each production ***/

  switch(Target)
  {
    case TARGET_WIN31:
        Resource = WIN31_OUT_RESOURCE_PROD_LEFT;
        break;
    case TARGET_UNIX:
        Resource = UNIX_OUT_RESOURCE_PROD_LEFT;
        break;
  }

  LoadString(hDLLInst,Resource,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  for(i = 0; i < nProduction; i++)  // Loop on productions
  {
    if(i % 10 == 0) WSOUTCLOSE(hf,"\r\n");

    if(i != nProduction - 1)
        wsprintf((LPSTR)OutBuffer,"%u,",hpProduction[i].Left);
    else
        wsprintf((LPSTR)OutBuffer,"%u",hpProduction[i].Left);

    WSOUTCLOSE(hf,OutBuffer);
  }

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n");
        break;
  }


  /*** Length of each production ***/

  switch(Target)
  {
    case TARGET_WIN31:
        Resource = WIN31_OUT_RESOURCE_PROD_LENGTH;
        break;
    case TARGET_UNIX:
        Resource = UNIX_OUT_RESOURCE_PROD_LENGTH;
        break;
  }

  LoadString(hDLLInst,Resource,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  for(i = 0; i < nProduction; i++)  // Loop on productions
  {
    if(i % 10 == 0)   WSOUTCLOSE(hf,"\r\n");

    if(i != nProduction - 1)
        wsprintf((LPSTR)OutBuffer,"%u,",hpProduction[i].Length);
    else
        wsprintf((LPSTR)OutBuffer,"%u",hpProduction[i].Length);

    WSOUTCLOSE(hf,OutBuffer);
  }

  switch(Target)
  {
    case TARGET_WIN31:
        WSOUTCLOSE(hf,"\r\nEND\r\n");
        break;
    case TARGET_UNIX:
        WSOUTCLOSE(hf,"\r\n};\r\n");
        break;
  }

  _lclose(hf);

  return TRUE;
}

/* Generation of the action - goto table */
BOOL PhoenixOutputActionGotoTable(LPSTR Directory)
{
  register WORD i,j;
  register DWORD lx;
  HFILE hf;
  OFSTRUCT ofs;

  lstrcpy((LPSTR)FileName,Directory);
  lstrcat((LPSTR)FileName,"\\wsparse.h");               // Parser tables header file

  for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i]; // Copy FileName to parameter
  LexErrorStringParam[i] = 0;   // Truncate to MAX_ERROR_PARAM

  hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
  if(hf == HFILE_ERROR)
  {
    LexErrorCode = LEXERR_CANT_CREATE_FILE;
    return FALSE;
  }

  LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);

  LoadString(hDLLInst,OUT_RESOURCE_ACTION,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
  WSOUTCLOSE(hf,OutBuffer);
 
  LoadString(hDLLInst,OUT_RESOURCE_ACTION_COUNT,(LPSTR)OutParam,MAX_OUT_BUFFER);
  wsprintf((LPSTR)OutBuffer,(LPSTR)OutParam,nCsi);
  WSOUTCLOSE(hf,OutBuffer);

  LoadString(hDLLInst,OUT_RESOURCE_TABLE_NAME,(LPSTR)OutParam,MAX_OUT_BUFFER);
  wsprintf((LPSTR)OutBuffer,(LPSTR)OutParam,nCsi);
  WSOUTCLOSE(hf,OutBuffer);

  wsprintf((LPSTR)OutBuffer,"\tswitch(State)\r\n\t{\r\n");
  WSOUTCLOSE(hf,OutBuffer);

  for(i = 0; i < nCsi; i++)
  {
    wsprintf((LPSTR)OutBuffer,"\t\tcase %u:\r\n\t\t\tswitch(Symbol)\r\n\t\t\t{\r\n",i);
    WSOUTCLOSE(hf,OutBuffer);
    lx = (DWORD)nSymbol * (DWORD)i;

    for(j = 0; j < nSymbol; j++)
    {
        if(hpActionGoto[lx + (DWORD)j] == BLANK) continue;
        if(hpActionGoto[lx + (DWORD)j] == ACCEPT)
            wsprintf((LPSTR)OutBuffer,"\t\t\t\tcase %u: Value = ACCEPT; break;\r\n",j);
        else
            wsprintf((LPSTR)OutBuffer,"\t\t\t\tcase %u: Value = %d; break;\r\n",j,hpActionGoto[lx + (DWORD)j]);
        WSOUTCLOSE(hf,OutBuffer);
    }
    wsprintf((LPSTR)OutBuffer,"\t\t\t\tdefault: Value = BLANK; break;\r\n\t\t\t}\r\n\t\t\tbreak;\r\n",j,BLANK);
    WSOUTCLOSE(hf,OutBuffer);
  }

  wsprintf((LPSTR)OutBuffer,"\t\tdefault: Value = BLANK; break;\r\n\t}\r\n\r\n\treturn Value;\r\n}\r\n");
  WSOUTCLOSE(hf,OutBuffer);

  _lclose(hf);

  return TRUE;
}

BOOL PhoenixOutputLexActions(LPSTR Directory,LPSTR LexFile, BOOL LineDirectives)
{
    WORD i,j,k;
    HFILE hf;
    OFSTRUCT ofs;

    lstrcpy((LPSTR)FileName,Directory);
    lstrcat((LPSTR)FileName,"\\wslexac.h");             // Header file containing lex actions

    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_CREATE_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    WSOUTCLOSE(hf,OutBuffer);

    for(i = 1; i <= nToken; i++)
    {
        /* Filter out the simple actions "return TOKEN_VALUE;" */
        if(hpP2RegExp[i-1].Action.CodeSize > 0)
        {
            wsprintf((LPSTR)OutBuffer,"\t\tcase %u:\r\n",i);

            WSOUTCLOSE(hf,OutBuffer);

            if(LineDirectives)
            {
                wsprintf((LPSTR)OutBuffer,"#line %lu \"",hpP2RegExp[i-1].Action.LineNumber);
                j = 0; while(OutBuffer[j]) j++;
                k = 0;
                while(LexFile[k])
                {
                    OutBuffer[j + k] = LexFile[k];
                    if(LexFile[k] == '\\') {j++; OutBuffer[j + k] = '\\';}
                    k++;
                }
                OutBuffer[j + k] = 0;
                lstrcat((LPSTR)OutBuffer,"\"\r\n\t\t\t");
                WSOUTCLOSE(hf,OutBuffer);
            }

            if(_lwrite(hf,hpP2RegExp[i-1].Action.hpCodeStart + 1,(UINT)hpP2RegExp[i-1].Action.CodeSize - 2)== HFILE_ERROR)
            {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

            wsprintf((LPSTR)OutBuffer,"\r\n\t\t\tbreak;\r\n");
            WSOUTCLOSE(hf,OutBuffer);
        }
    }

    _lclose(hf);

    return TRUE;
}

BOOL PhoenixOutputParserActions(LPSTR Directory,LPSTR ParseFile, BOOL LineDirectives)
{
    WORD i,j,k;
    HFILE hf;
    OFSTRUCT ofs;

    lstrcpy((LPSTR)FileName,Directory);
    lstrcat((LPSTR)FileName,"\\wsparac.h");             // Header file containing parser actions

    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_CREATE_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    WSOUTCLOSE(hf,OutBuffer);

    for(i = 0; i < nProduction; i++)
    {
        if(hpProduction[i].hpActionStart == NULL) continue;     // No action on this production

        wsprintf((LPSTR)OutBuffer,"\t\tcase %u:\r\n",i);

        WSOUTCLOSE(hf,OutBuffer);

        if(LineDirectives)
        {
            wsprintf((LPSTR)OutBuffer,"#line %lu \"",hpProduction[i].ActionLineNumber);
            j = 0; while(OutBuffer[j]) j++;
            k = 0;
            while(ParseFile[k])
            {
                OutBuffer[j + k] = ParseFile[k];
                if(ParseFile[k] == '\\') {j++; OutBuffer[j + k] = '\\';}
                k++;
            }
            OutBuffer[j + k] = 0;
            lstrcat((LPSTR)OutBuffer,"\"\r\n\t\t\t");
            WSOUTCLOSE(hf,OutBuffer);
        }

        if(_lwrite(hf,hpProduction[i].hpActionStart + 1,(UINT)hpProduction[i].ActionSize - 2)== HFILE_ERROR)
        {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

        wsprintf((LPSTR)OutBuffer,"\r\n\t\t\tbreak;\r\n");
        WSOUTCLOSE(hf,OutBuffer);
    }

    _lclose(hf);

    return TRUE;
}

BOOL PhoenixOutputParserTokens(LPSTR Directory)
{
    WORD i;
    HFILE hf;
    OFSTRUCT ofs;
    unsigned char c;

    lstrcpy((LPSTR)FileName,Directory);
    lstrcat((LPSTR)FileName,"\\wstoken.h");             // Header file containing parser tokens

    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_CREATE_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    WSOUTCLOSE(hf,OutBuffer);

    wsprintf((LPSTR)OutBuffer,"/* Token definitions */\r\n\r\n");
    WSOUTCLOSE(hf,OutBuffer);

    for(i = 2; i < nTerm; i++)                                  // Loop on terminals: skip endmarker and error token
    {
    	char tmpBuf[MAX_OUT_BUFFER + 1];
    	
        c = hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize];
        hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize] = 0;
        CopyHugeString(tmpBuf, hpSymbol[i].hpSymbolStart);
        wsprintf((LPSTR)OutBuffer,"#define %-25s %u\r\n",(LPSTR)tmpBuf,i);
        WSOUTCLOSE(hf,OutBuffer);
    }

    _lclose(hf);

    return TRUE;
}

BOOL PhoenixOutputLexRoutines(LPSTR Directory, LPSTR LexFile, BOOL LineDirectives)
{
    register WORD i,j,k;
    HFILE hf;
    OFSTRUCT ofs;

    lstrcpy((LPSTR)FileName,Directory);
    lstrcat((LPSTR)FileName,"\\wslexrt.h");             // Header file containing user support routines

    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_CREATE_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    WSOUTCLOSE(hf,OutBuffer);

    wsprintf((LPSTR)OutBuffer,"#include \"wstoken.h\"\r\n");
    WSOUTCLOSE(hf,OutBuffer);

    switch(Target)
    {
        case TARGET_WIN31:
            break;
        case TARGET_UNIX:
            wsprintf((LPSTR)OutBuffer,"#ifndef PH_NO_SCAN_TABLES\r\n#include \"winscant.h\"\r\n#endif\r\n\r\n");
            WSOUTCLOSE(hf,OutBuffer);
            break;
    }

    wsprintf((LPSTR)OutBuffer,"#define DFA_START\t%u\r\n\r\n",DfaStart);
    WSOUTCLOSE(hf,OutBuffer);

    if(LineDirectives)
    {
        wsprintf((LPSTR)OutBuffer,"#line %lu \"",P1Code.LineNumber);
        j = 0; while(OutBuffer[j]) j++;
        k = 0;
        while(LexFile[k])
        {
            OutBuffer[j + k] = LexFile[k];
            if(LexFile[k] == '\\') {j++; OutBuffer[j + k] = '\\';}
            k++;
        }
        OutBuffer[j + k] = 0;
        lstrcat((LPSTR)OutBuffer,"\"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
    }

    if(_lwrite(hf,P1Code.hpCodeStart + 1,(UINT)P1Code.CodeSize - 2)== HFILE_ERROR)
    {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

    WSOUTCLOSE(hf,"\r\n\r\n");

    if(LineDirectives)
    {
        wsprintf((LPSTR)OutBuffer,"#line %lu \"",P3Code.LineNumber);
        j = 0; while(OutBuffer[j]) j++;
        k = 0;
        while(LexFile[k])
        {
            OutBuffer[j + k] = LexFile[k];
            if(LexFile[k] == '\\') {j++; OutBuffer[j + k] = '\\';}
            k++;
        }
        OutBuffer[j + k] = 0;
        lstrcat((LPSTR)OutBuffer,"\"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
    }

    if(_lwrite(hf,P3Code.hpCodeStart,(UINT)P3Code.CodeSize)== HFILE_ERROR)
    {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

    WSOUTCLOSE(hf,"\r\n\r\n");

    _lclose(hf);

    return TRUE;
}

BOOL PhoenixOutputParserRoutines(LPSTR Directory, LPSTR ParserFile, BOOL LineDirectives)
{
    register WORD i,j,k;
    HFILE hf;
    OFSTRUCT ofs;

    lstrcpy((LPSTR)FileName,Directory);
    lstrcat((LPSTR)FileName,"\\wsparrt.h");             // Header file containing user support routines

    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_CREATE_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    WSOUTCLOSE(hf,OutBuffer);

    wsprintf((LPSTR)OutBuffer,"/* Generated file for parser declarations and support routines */\r\n\r\n");
    WSOUTCLOSE(hf,OutBuffer);

    if(LineDirectives)
    {
        wsprintf((LPSTR)OutBuffer,"#line %lu \"",P1Code.LineNumber);
        j = 0; while(OutBuffer[j]) j++;
        k = 0;
        while(ParserFile[k])
        {
            OutBuffer[j + k] = ParserFile[k];
            if(ParserFile[k] == '\\') {j++; OutBuffer[j + k] = '\\';}
            k++;
        }
        OutBuffer[j + k] = 0;
        lstrcat((LPSTR)OutBuffer,"\"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
    }

    if(_lwrite(hf,P1Code.hpCodeStart + 1,(UINT)P1Code.CodeSize - 2)== HFILE_ERROR)
    {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

    WSOUTCLOSE(hf,"\r\n\r\n");

    if(LineDirectives)
    {
        wsprintf((LPSTR)OutBuffer,"#line %lu \"",P3Code.LineNumber);
        j = 0; while(OutBuffer[j]) j++;
        k = 0;
        while(ParserFile[k])
        {
            OutBuffer[j + k] = ParserFile[k];
            if(ParserFile[k] == '\\') {j++; OutBuffer[j + k] = '\\';}
            k++;
        }
        OutBuffer[j + k] = 0;
        lstrcat((LPSTR)OutBuffer,"\"\r\n");
        WSOUTCLOSE(hf,OutBuffer);
    }

    if(_lwrite(hf,P3Code.hpCodeStart,(UINT)P3Code.CodeSize)== HFILE_ERROR)
    {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

    WSOUTCLOSE(hf,"\r\n\r\n");

    _lclose(hf);

    return TRUE;
}

BOOL PhoenixOutputLOptions(LPSTR Directory,BOOL Pack)
{
    register WORD i;
    HFILE hf;
    OFSTRUCT ofs;

    lstrcpy((LPSTR)FileName,Directory);
    lstrcat((LPSTR)FileName,"\\loptions.h");                // Header file containing compile options

    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_CREATE_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    WSOUTCLOSE(hf,OutBuffer);

    wsprintf((LPSTR)OutBuffer,"/* Generated file for compile options */\r\n\r\n");
    WSOUTCLOSE(hf,OutBuffer);

    switch(Target)
    {
        case TARGET_WIN31:
            wsprintf((LPSTR)OutBuffer,"#define PHK_WIN31\r\n");
            break;
        case TARGET_UNIX:
            wsprintf((LPSTR)OutBuffer,"#define PHK_UNIX\r\n");
            break;
    }
    WSOUTCLOSE(hf,OutBuffer);

    if(Pack)
    {
        wsprintf((LPSTR)OutBuffer,"#define PACKED_ARRAYS\r\n");
        WSOUTCLOSE(hf,OutBuffer);
    }

    _lclose(hf);

    return TRUE;
}

BOOL PhoenixOutputPOptions(LPSTR Directory,BOOL Pack)
{
    register WORD i;
    HFILE hf;
    OFSTRUCT ofs;

    lstrcpy((LPSTR)FileName,Directory);
    lstrcat((LPSTR)FileName,"\\poptions.h");                // Header file containing compile options

    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(FileName, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_CREATE_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    WSOUTCLOSE(hf,OutBuffer);

    wsprintf((LPSTR)OutBuffer,"/* Generated file for compile options */\r\n\r\n");
    WSOUTCLOSE(hf,OutBuffer);

    switch(Target)
    {
        case TARGET_WIN31:
            wsprintf((LPSTR)OutBuffer,"#define PHK_WIN31\r\n");
            break;
        case TARGET_UNIX:
            wsprintf((LPSTR)OutBuffer,"#define PHK_UNIX\r\n");
            break;
    }
    WSOUTCLOSE(hf,OutBuffer);

    if(Pack)
    {
        wsprintf((LPSTR)OutBuffer,"#define PACKED_ARRAYS\r\n");
        WSOUTCLOSE(hf,OutBuffer);
    }

    wsprintf((LPSTR)OutBuffer,"#define NONTERMINAL_START\t%u\r\n",nTerm);
    WSOUTCLOSE(hf,OutBuffer);

    _lclose(hf);

    return TRUE;
}

BOOL __export HOST_EXPORT_QUALIFIER PhoenixAnalyzeLexFile(
    HWND    hwnd,           // Handle of parent window
    LPSTR   LexFile,        // Input lex source file
    LPSTR   TargetDir,      // Target directory
    int     TargetPlatform, // Target platform: TARGET_WIN31 or TARGET_UNIX,
    BOOL    LineDirectives, // Generate #line directives ?
    BOOL    Pack,           // Pack final DFA ?
    WORD    PPercent        // Percentage of matching gotos in two states, over which decision is taken to
                            // merge the states
    )
{
  SourceType = SOURCE_LEXICAL;
  Target = TargetPlatform;
  LexErrorCode = LEXERR_NO_ERROR;
  Packed = Pack;

#ifdef PH_TRACE
  if(*TraceFile)
  {
    WORD i;

    for(i = 0; i < MAX_FILE_NAME && TraceFile[i]; i++) LexTraceFile[i] = TraceFile[i];  // Copy Trace File Name
    LexTraceFile[i] = 0;    // Truncate to MAX_FILE_NAME
    LexTrace = TRUE;
  }
  else LexTrace = FALSE;
#else
  LexTrace = FALSE;
#endif

  LexInit();
  if(LoadFile(LexFile) == FALSE) {LexError(hwnd); LexEnd(); return FALSE;}
  if(LexPrepare() == FALSE) {LexError(hwnd); LexEnd(); return FALSE;}
  if(ConstructDfa() == FALSE) {LexError(hwnd); LexEnd(); DFAFree(); return FALSE;}

  if(Pack)
  {
    Percent = PPercent;
    if(PackDFA() == FALSE) {LexError(hwnd); LexEnd(); DFAFree(); return FALSE;}
    if(PhoenixOutputPackedTables(TargetDir) == FALSE) {LexError(hwnd); LexEnd(); DFAFree(); return FALSE;}
  }
  else
    if(PhoenixOutputLexTables(TargetDir) == FALSE) {LexError(hwnd); LexEnd(); DFAFree(); return FALSE;}

  if(PhoenixOutputLexActions(TargetDir, LexFile, LineDirectives) == FALSE) {LexError(hwnd); LexEnd(); DFAFree(); return FALSE;}
  if(PhoenixOutputLexRoutines(TargetDir, LexFile, LineDirectives) == FALSE) {LexError(hwnd); LexEnd(); DFAFree(); return FALSE;}
  if(PhoenixOutputLOptions(TargetDir,Pack) == FALSE) {LexError(hwnd); LexEnd(); DFAFree(); return FALSE;}
  LexEnd();     // Free intermediate structures for regular expressions
  DFAFree();    // Free arrays for DFA construction

  return TRUE;
}

BOOL __export HOST_EXPORT_QUALIFIER PhoenixAnalyzeParserFile(
    HWND    hwnd,           // Handle of parent window
    LPSTR   ParserFile,     // Input parser source file
    LPSTR   TargetDir,      // Target directory
    int     TargetPlatform, // Target platform: TARGET_WIN31 or TARGET_UNIX
    BOOL    LineDirectives, // Generate #line directives ?
    BOOL    Pack            // Pack Action and Goto tables ?
    )
{
  SourceType = SOURCE_PARSER;
  Target = TargetPlatform;
  LexErrorCode = LEXERR_NO_ERROR;

#ifdef PH_TRACE
  if(*TraceFile)
  {
    WORD i;

    for(i = 0; i < MAX_FILE_NAME && TraceFile[i]; i++) LexTraceFile[i] = TraceFile[i];  // Copy Trace File Name
    LexTraceFile[i] = 0;    // Truncate to MAX_FILE_NAME
    LexTrace = TRUE;
  }
  else LexTrace = FALSE;
#else
  LexTrace = FALSE;
#endif

  PInit();      // Initialization

  if(LoadFile(ParserFile) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
  if(ParserPrepare() == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
  if(BuildParsingTable(TargetDir) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}

  if(PhoenixOutputProductions(TargetDir) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
  if(Pack)
  {
    if(PackActionGoto() == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
    if(PhoenixOutputPackedActionGoto(TargetDir) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
  }
  else
    if(PhoenixOutputActionGotoTable(TargetDir) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}

  if(PhoenixOutputParserActions(TargetDir, ParserFile, LineDirectives) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
  if(PhoenixOutputParserRoutines(TargetDir, ParserFile, LineDirectives) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
  if(PhoenixOutputParserTokens(TargetDir) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}
  if(PhoenixOutputPOptions(TargetDir,Pack) == FALSE) {LexError(hwnd); PEnd(); return FALSE;}

  PEnd();       // Free arrays for parser tables construction

  if(ConsultReport)
    MessageBox(hwnd,"Error(s),conflict(s) or warning(s) detected in input grammar. Consult report file.",
                APPNAME,MB_OK|MB_ICONSTOP);

  return TRUE;
}

/* Manage Phoenix errors */
void LexError(HWND hwnd)
{
  static unsigned char ErrorMessageParam[MAX_ERROR_MESSAGE+1];
  static unsigned char ErrorMessage[MAX_ERROR_MESSAGE+1];
  static unsigned char ErrorLine[50];
  WORD charcode;
  BOOL NoLine;

  NoLine = FALSE;

  switch(LexErrorCode)
  {
    /*** Parser generator errors ***/

    case PERR_NDESC_TOO_LOW:
    case PERR_NTERM_TOO_SHORT:
    case PERR_COUPLES_TOO_SHORT:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessage,MAX_ERROR_MESSAGE);
        NoLine = TRUE;
        break;
    case PERR_TOKEN_MULT_DECLARED:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessageParam,MAX_ERROR_MESSAGE);
        wsprintf((LPSTR)ErrorMessage,(LPSTR)ErrorMessageParam,(LPSTR)LexErrorStringParam);
        NoLine = TRUE;
        break;

    /*** Lexical analyzer generator errors ***/

    case LEXERR_REACHED_EOF_COMMENT:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessage,MAX_ERROR_MESSAGE);
        LexErrorLine = LexErrorLineComment;
        LexErrorColumn = LexErrorColComment;
        break;
    case LEXERR_REACHED_EOF_STRING:
    case LEXERR_REACHED_EOL_STRING:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessage,MAX_ERROR_MESSAGE);
        LexErrorLine = LexErrorLineString;
        LexErrorColumn = LexErrorColString;
        break;
    case LEXERR_REACHED_EOF_CODE:
    case LEXERR_REACHED_EOF_CLASS:
    case LEXERR_REACHED_EOL_CLASS:
    case LEXERR_REACHED_EOF_ID:
    case PERR_INCOMPLETE_TOKEN:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessage,MAX_ERROR_MESSAGE);
        break;  // LexErrorLine and LexErrorColumn already set
    case LEXERR_REG_EXP_TOO_LARGE:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessageParam,MAX_ERROR_MESSAGE);
        wsprintf((LPSTR)ErrorMessage,(LPSTR)ErrorMessageParam,LexErrorIntParam);
        LexErrorLine = yyline;
        LexErrorColumn = yycol;
        break;
    case LEXERR_INVALID_CHARACTER:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessageParam,MAX_ERROR_MESSAGE);
        charcode = (WORD)LexErrorCharParam;
        if(LexErrorCharParam < ' ') LexErrorCharParam = ' ';    // Non printable characters
        wsprintf((LPSTR)ErrorMessage,(LPSTR)ErrorMessageParam,LexErrorCharParam,charcode);
        LexErrorLine = yyline;
        LexErrorColumn = yycol;
        break;
    case LEXERR_ID_NOT_DECLARED:
    case LEXERR_CANT_OPEN_FILE:
    case LEXERR_CANT_READ_FILE:
    case LEXERR_CANT_CREATE_FILE:
    case LEXERR_CANT_WRITE_FILE:
    case LEXERR_CANT_EOF_FILE:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessageParam,MAX_ERROR_MESSAGE);
        wsprintf((LPSTR)ErrorMessage,(LPSTR)ErrorMessageParam,(LPSTR)LexErrorStringParam);
        LexErrorLine = yyline;
        LexErrorColumn = yycol;
        NoLine = TRUE;
        break;
    default:
        LoadString(hDLLInst,LexErrorCode,(LPSTR)ErrorMessage,MAX_ERROR_MESSAGE);
        LexErrorLine = yyline;
        LexErrorColumn = yycol;
        break;
  }

  if(NoLine)
    MessageBox(hwnd,(LPSTR)ErrorMessage,APPNAME,MB_OK|MB_ICONSTOP);
  else
  {
      if(SourceType == SOURCE_LEXICAL)
        wsprintf((LPSTR)ErrorLine,"%s: Error at line %ld, column %ld in lexical source",(LPSTR)APPNAME,LexErrorLine,LexErrorColumn);
      else          // SOURCE_PARSER
        wsprintf((LPSTR)ErrorLine,"%s: Error at line %ld, column %ld in parser source",(LPSTR)APPNAME,LexErrorLine,LexErrorColumn);
    MessageBox(hwnd,(LPSTR)ErrorMessage,(LPSTR)ErrorLine,MB_OK|MB_ICONSTOP);
  }
}
