/******************************************************************
* Phoenix Lexical Analyzer Generator.                             *
* Input buffer reading, and storing into intermediate structures. *
* One difficulty is due to the management of CR + LF characters   *
* at the end of each line.                                        *
******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include <io.h>
#include "ph_scan.h"
#include "ph_error.h"

#define PERCENT     55      // Percentage of matching GOTOs in 2 states, over which the decision is taken to
                            // merge the states (for packed arrays only)


/* Initialize lex variables */
BOOL LexInit()
{
  hpInputBuffer         = NULL;
  hpClass               = NULL;
  if(AllocateClass() == FALSE) return FALSE;
  lpP1Decl              = NULL;
  hpP2RegExp            = NULL;
  P1Code.hpCodeStart    = 0;
  P1Code.CodeSize       = 0;
  P3Code.hpCodeStart    = 0;
  P3Code.CodeSize       = 0;
  LexErrorCode          = LEXERR_NO_ERROR;
  yyline                = 1;
  yycol                 = 1;
  CaseSensitive         = TRUE;
  PBase                 = NULL;
  PDefault              = NULL;
  PNext                 = NULL;
  PCheck                = NULL;

  return TRUE;
}

/* Memory management to do before exiting program */
BOOL LexEnd()
{
  if(hpInputBuffer != NULL)
  {
    GlobalFreePtr(hpInputBuffer);
    hpInputBuffer = NULL;
  }

  if(lpP1Decl != NULL)
  {
    GlobalFreePtr(lpP1Decl);
    lpP1Decl = NULL;
  }

  if(hpP2RegExp != NULL)
  {
    GlobalFreePtr(hpP2RegExp);
    hpP2RegExp = NULL;
  }

  if(hpClass != NULL)
  {
    GlobalFreePtr(hpClass);
    hpClass = NULL;
  }

  if(PBase != NULL)
  {
    GlobalFreePtr(PBase);
    PBase = NULL;
  }

  if(PDefault != NULL)
  {
    GlobalFreePtr(PDefault);
    PDefault = NULL;
  }

  if(PNext != NULL)
  {
    GlobalFreePtr(PNext);
    PNext = NULL;
  }

  if(PCheck != NULL)
  {
    GlobalFreePtr(PCheck);
    PCheck = NULL;
  }

  return TRUE;
}

/* Read the input file; store it into the input buffer */
BOOL LoadFile(LPSTR FileName)
{
  register WORD i;
  HFILE hf;
  OFSTRUCT ofs;
  long count;

  hf = OpenFile(FileName, &ofs, OF_READ);
  if (hf == HFILE_ERROR)
  {
    LexErrorCode = LEXERR_CANT_OPEN_FILE;
    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM
    return FALSE;
  }

  /* Get the length of the file in bytes */
  if((InputSize = _filelength(hf)) == -1L)
  {
    _lclose(hf);
    LexErrorCode = LEXERR_INVALID_FILE_HANDLE;
    return FALSE;
  }

  /* Read file into buffer */
  hpInputBuffer = GlobalAllocPtr(GHND,(DWORD)(InputSize + 1));
  if(hpInputBuffer == NULL)
  {
    _lclose(hf);
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }
  count = _hread(hf,hpInputBuffer,InputSize);
  if(count == -1L)
  {
    _lclose(hf);
    GlobalFreePtr(hpInputBuffer);
    hpInputBuffer = NULL;
    LexErrorCode = LEXERR_CANT_READ_FILE;
    for(i = 0; i < MAX_ERROR_PARAM && FileName[i]; i++) LexErrorStringParam[i] = FileName[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM
    return FALSE;
  }
  InputSize = count;    // Real buffer size
  hpInputBuffer[InputSize] = 0; // Trailing zero
  _lclose(hf);

  return TRUE;
}

/* Reads a C comment begining with slash-star, or double-slash. Returns an index to the last
   character of the comment (slash or end-of-line). */
BOOL ReadComment(LPHUGE hpComment, LPDWORD lpIndex)
{
  register WORD i;
  BOOL DoubleSlash;

  if(*hpComment != '/' || hpComment[1] != '/' && hpComment[1] != '*') {LexErrorCode = LEXERR_INVALID_COMMENT; return FALSE;}
  if(hpComment[1] == '/') DoubleSlash = TRUE; else DoubleSlash = FALSE;
  LexErrorLineComment = yyline;
  LexErrorColComment = yycol;
  i = 2; yycol += 2;
  while(1)
  {
    switch(hpComment[i])
    {
      case 0:   if(DoubleSlash == FALSE)
                {LexErrorCode = LEXERR_REACHED_EOF_COMMENT; return FALSE;}
                else {*lpIndex = i; return TRUE;}
      case 10:  yyline++; yycol = 1; if(DoubleSlash) {*lpIndex = i; return TRUE;}
                break;
      case '*': if(DoubleSlash) break;
                if(hpComment[i+1] == '/') {*lpIndex = i+1; yycol++; return TRUE;}
      default:  break;
    }
    if(hpComment[i] != 10 && hpComment[i] != 13) yycol++;
    i++;
  }
}

/* Reads a C string (enclosed in double quotes). Returns an index to the last double quote. */
BOOL ReadString(LPHUGE hpString, LPDWORD lpIndex)
{
  register WORD i;

  if(*hpString != '"') {LexErrorCode = LEXERR_INVALID_STRING; return FALSE;}
  LexErrorLineString = yyline;
  LexErrorColString = yycol;
  i = 1; yycol++;
  while(1)
  {
    switch(hpString[i])
    {
      case 0:       LexErrorCode = LEXERR_REACHED_EOF_STRING; return FALSE;
      case 10:      yyline++; yycol = 1; break;
      case '\\':    if(hpString[i+1] == 0) {LexErrorCode = LEXERR_REACHED_EOF_STRING; return FALSE;}
                    yycol++; i++; break;
      case '"':     *lpIndex = i; return TRUE;
      default:      break;
    }
    if(hpString[i] != 10 && hpString[i] != 13) yycol++;
    i++;
  }
}

/* Reads C code between { and }, then returns an index to final }. hpCode must initially point to {. */
BOOL ReadCode(LPHUGE hpCode, LPDWORD lpIndex)
{
  register DWORD i;
  int level;

  if(*hpCode != '{') {LexErrorCode = LEXERR_MISSING_LEFT_BRACE; return FALSE;}
  LexErrorLine = yyline;
  LexErrorColumn = yycol;
  level = 0;
  i = 0;
  while(1)
  {
    switch(hpCode[i])
    {
      case '}':     level--; if(level > 0) break; else {*lpIndex = i; return TRUE;}
      case '{':     level++; break;
      case '/':     if(hpCode[i+1] == '/' || hpCode[i+1] == '*')
                    {
                        if(ReadComment(hpCode+i,lpIndex) == FALSE) return FALSE;
                        i += *lpIndex;
                    }
                    break;
      case '"':     if(ReadString(hpCode+i,lpIndex) == FALSE) return FALSE;
                    i += *lpIndex; break;
      case '\'':    i++; yycol++;
                    if(hpCode[i] == '\\')   // Beware of '\''
                    {
                        if(hpCode[i+1] == 0) {LexErrorCode = LEXERR_REACHED_EOF_BSLASH; return FALSE;}
                        else
                        {
                            if(hpCode[i+1] == 13) yycol++;
                            else yycol += 2;
                            i += 2;
                        }
                    }
                    while(hpCode[i] && hpCode[i] != 10 && hpCode[i] != '\'') {if(hpCode[i] != 13) yycol++; i++;}
                    if(hpCode[i] == 10) {LexErrorCode = LEXERR_REACHED_EOL_CHAR; return FALSE;}
                    if(hpCode[i] == 0) {LexErrorCode = LEXERR_REACHED_EOF_CHAR; return FALSE;}
                    else break;
      case 0:       LexErrorCode = LEXERR_REACHED_EOF_CODE; return FALSE;
      case 10:      yyline++; yycol = 1; break;
      default:      break;
    }
    if(hpCode[i] != 10 && hpCode[i] != 13) yycol++;
    i++;
  }
}

/* Reads an identifier. Returns an index to the first character after it. */
BOOL ReadId(LPHUGE hpId, LPDWORD lpIndex)
{
  register WORD i;

  if(IsCharAlpha(*hpId)== FALSE && *hpId != '_') {LexErrorCode = LEXERR_INVALID_IDENTIFIER; return FALSE;}
  i = 1; yycol++;
  while(hpId[i] && (IsCharAlphaNumeric(hpId[i]) || hpId[i] == '_')) {i++; yycol++;}
  *lpIndex = i;
  return TRUE;
}

/* Reads whitespace characters, and comments, and returns an index to the first nonwhitespace.
   If ReadEOL = FALSE, then characters 10 and 13 are NOT counted as whitespace. */
BOOL ReadWhiteSpace(LPHUGE hpWhiteSpace, BOOL ReadEOL, LPDWORD lpIndex)
{
  register DWORD i;

  i = 0;
  while(1)
  {
    switch(hpWhiteSpace[i])
    {
      case ' ':
      case '\t':    break;
      case 13:      if(ReadEOL == FALSE) {*lpIndex = i; return TRUE;} else break;
      case 10:      yyline++; yycol = 1; if(ReadEOL == FALSE) {*lpIndex = i; return TRUE;}
                    break;
      case '/':     if(hpWhiteSpace[i+1] == '/' || hpWhiteSpace[i+1] == '*')
                    {
                        if(ReadComment(hpWhiteSpace+i,lpIndex) == FALSE) return FALSE;
                        i += *lpIndex;
                    }
                    else {*lpIndex = i; return TRUE;}
                    break;
      default:      *lpIndex = i; return TRUE;
    }
    if(hpWhiteSpace[i] != 10 && hpWhiteSpace[i] != 13) yycol++;
    i++;
  }
}

/* Replace the ID at position hp1, of length idlen, by it's definition (processing Part 2).
   Write the definition to array lpResult, with additional surrounding parenthesis, if needed.
   New position in the array is returned via NewPos. */
BOOL ReplaceId(LPHUGE hp1, WORD idlen, LPWORD lpResult, DWORD LimitPos, DWORD *NewPos)
{
  register WORD i,j;
  WORD base,level;
  BOOL Found,Escape;
  LPLEXP1DECL lp;
  WORD c,c1,work[2];

  for(i = 0; i < nP1Decl; i++)
  {
    lp = lpP1Decl+i;
    if(lp -> IdSize != idlen) continue;
    Found = TRUE;
    work[1] = 0;
    for(j = 0; j < lp -> IdSize; j++)
    {
      work[0] = lp -> lp1IdStart[j];
      AnsiUpperBuff((LPSTR)work,1);
      c = work[0];

      work[0] = hp1[j];
      AnsiUpperBuff((LPSTR)work,1);
      c1 = work[0];
      if(c != c1) {Found = FALSE; break;}   // Case insensitive compare!!
    }
    if(Found == FALSE) continue;

    /*** Found declaration in Part 1: operate replacement. IMPORTANT: add surrounding parenthesis if needed!! ***/

    if(lp -> RegExp[0] == '[' && lp -> RegExp[lp -> RegExpSize - 1] == ']')
    {
        base = 1;
        level = 0;
        for(j = 0; j < lp -> RegExpSize; j++)
        {
            switch(lp -> RegExp[j])
            {
                case '[':
                    if(Escape == FALSE) level++; else Escape = FALSE;
                    break;
                case ']':
                    if(Escape == FALSE)
                    {
                        if(j == lp -> RegExpSize - 1)
                        {
                            base = 0;   // No parenthesis if surrounding square brackets
                        }
                        else
                        {
                            level--;
                            if(level == 0) goto OUT_LOOP;   // Mandatory surrounding parenthesis
                        }
                    }
                    else Escape = FALSE;
                    break;
                case '\\': Escape = TRUE; break;
                default: if(Escape) Escape = FALSE;
                    break;
            }
        }
        OUT_LOOP: ;
    }
    else
    {
        c = lp -> RegExp[lp -> RegExpSize - 1];
        if(c == '?' || c == '*' || c == '+') base = 1;  // Parenthesis when ?, * or + at the end
        else
        {
            base = 1;
            if(lp -> RegExp[0] == '(' && lp -> RegExp[lp -> RegExpSize - 1] == ')')
            {
                level = 0;
                for(j = 0; j < lp -> RegExpSize; j++)
                {
                    if(lp -> RegExp[j] == '(') level++;
                    else
                        if(lp -> RegExp[j] == ')')
                        {
                            if(j == lp -> RegExpSize - 1)
                            {
                                if(level == 1) base = 0;    // No parenthesis if surrounding parenthesis already
                            }
                            else
                            {
                                level--;
                                if(level == 0) break;       // Mandatory surrounding parenthesis
                            }
                        }
                }
            }

            if(base == 1)
            {
                base = 0;
                for(j = 0; j < lp -> RegExpSize; j++)
                {
                    if(lp -> RegExp[j] == '|') {base = 1; break;}   // Parenthesis if | in expression
                }
            }
        }
    }

    if(lp -> RegExpSize + 2 * base > LimitPos)
    {
        LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
        LexErrorIntParam = MAX_REG_EXP;
        return FALSE;
    }
    if(base) lpResult[0] = '(';
    for(j = 0; j < lp -> RegExpSize; j++) lpResult[j + base] = lp -> RegExp[j];
    if(base) lpResult[lp -> RegExpSize + 1] = ')';
    *NewPos = lp -> RegExpSize + 2 * base;
    return TRUE;
  }

  LexErrorCode = LEXERR_ID_NOT_DECLARED;
  for(i = 0; i < idlen && i < MAX_ERROR_PARAM; i++) LexErrorStringParam[i] = hp1[i];
  LexErrorStringParam[i] = 0;
  return FALSE;
}

/* Reads in the regular expression a sequence of characters begining by a backslash,
   evaluates it and returns the equivalent ANSI character code. */
unsigned char BackslashConstant(LPHUGE s, LPWORD Length)
{
    unsigned char value,c1,c2,c3,work[2];

    /* ASCII character in octal notation ? */
    c1 = s[1];      // Skip the backslash
    if(c1 >= '0' && c1 <= '7')
    {
        c2 = s[2];
        if(c2 >= '0' && c2 <= '7')
        {
            c3 = s[3];
            if(c3 >= '0' && c3 <= '7')
            {
                c1 -= '0';
                c2 -= '0';
                c3 -= '0';
                *Length = 4;
                return c1 * 64 + c2 * 8 + c3;   // Octal notation !!
            }
        }
    }

    /* Not an octal notation: other cases */
    *Length = 2;
    switch(c1)
    {
        case '0':   value = 0;    break;    // NUL character
        case 'r':   value = '\r'; break;    // Carriage return
        case 'n':   value = '\n'; break;    // (ANSI 10)
        case 't':   value = '\t'; break;    // Horizontal Tabulation
        case 'f':   value = '\f'; break;    // Form feed
        case 'x':   // ASCII character in hex notation ?
                    work[1] = 0;
                    work[0] = s[2];     // Skip the 'x'
                    AnsiUpperBuff((LPSTR)work,1);
                    c1 = work[0];
                    if(c1 >= '0' && c1 <= '9' || c1 >= 'A' && c1 <= 'F')
                    {
                        work[0] = s[3];
                        AnsiUpperBuff((LPSTR)work,1);
                        c2 = work[0];
                        if(c2 >= '0' && c2 <= '9' || c2 >= 'A' && c2 <= 'F')
                        {
                            if(c1 >= 'A' && c1 <= 'F') c1 = c1 - 'A' + 10;
                            else c1 -= '0';
                            if(c2 >= 'A' && c2 <= 'F') c2 = c2 - 'A' + 10;
                            else c2 -= '0';
                            value = c1 * 16 + c2;   // Hex notation !!
                            *Length = 4;
                            break;
                        }
                    }
                    // Intentionally no break instruction: fallthrough
        default:    value = s[1]; break;
    }

    return value;
}

/* Replaces a class by an array of BYTES of type ONE_CLASS. If this class is not already registered, the function
   registers it. Class code is returned via ClassCode pointer. Length of class specification is returned via Clength pointer. */
BOOL ExpandClass(LPHUGE lpR, LPWORD ClassCode, LPWORD Clength)
{
    register WORD i,j,RangeBegin,Index;
    unsigned char c;
    BOOL Complemented,FoundClass;
    ONE_CLASS Tab;
    static WORD Length;

    /*** Expand class ***/

    Index = 0;
    for(i = 1; i < MAX_CHARS; i++) Tab[i] = 0;
    c = lpR[Index];
    if(c == '.')
    {
        /* Any character but newline */
        for(i = 1; i < MAX_CHARS; i++)
            if(i != '\n') Tab[i] = 1;
    }
    else    // c is equal to '['
    {
        Index++;        // Skip the [
        RangeBegin = EOI;
        c = lpR[Index];
        if(c == '^')    // Complemented character class: matches all characters NOT in the class!
        {
            Complemented = TRUE;
            Index++;
        }
        else Complemented = FALSE;

        while((c = lpR[Index]) != ']')  // Loop on all characters of the class definition
        {
            if(c == EOI) {LexErrorCode = LEXERR_REG_EXP_SYNTAX; return FALSE;}
            if(c != '-'
               || c == '-' && lpR[Index + 1] == ']' // A simple character. A '-' before a ']' cannot
                                                    // represent a character range!
               || c == '-' && RangeBegin == EOI)    // A '-' just after a '[' cannot either!
            {
                if(c == '\\')
                {
                    /* Backslash in class: may escape as in C */
                    c = BackslashConstant(lpR + Index,(LPWORD)&Length);
                    Index += Length - 1;
                }

                Tab[c] = 1;
                RangeBegin = c;                                 // Store first character of a potential range
            }
            else                                                // '-': A character range
            {
                Index++;                                // Move to upper bound of range
                c = lpR[Index];
                if(c == '\\')
                {
                    /* Backslash in class: may escape as in C */
                    c = BackslashConstant(lpR + Index,(LPWORD)&Length);
                    Index += Length - 1;
                }

                if(c < RangeBegin)                              // Lower bound > upper bound
                {LexErrorCode = LEXERR_INVALID_RANGE; return FALSE;}
                for(i = RangeBegin; i <= c; i++) Tab[i] = 1;
                RangeBegin = EOI;                               // Reset Upper bound to none
            }
            Index++;
        }

        if(CaseSensitive == FALSE)
        {
            for(i = 'A'; i <= 'Z'; i++)
                if(Tab[i] == 1) Tab[i + 'a' - 'A'] = 1;
            for(i = 'a'; i <= 'z'; i++)
                if(Tab[i] == 1) Tab[i - 'a' + 'A'] = 1;
        }

        if(Complemented)
            for(i = 1; i < MAX_CHARS; i++)
                if(Tab[i] == 0) Tab[i] = 1; else Tab[i] = 0;
    }


    /*** Register class if does not exist yet ***/

    FoundClass = FALSE;
    for(i = 0; i < nClass; i++)
    {
        for(j = 0; j < MAX_CHARS; j++) if(hpClass[i][j] != Tab[j]) break;
        if(j == MAX_CHARS) {FoundClass = TRUE; break;}
    }

    if(FoundClass) *ClassCode = CLASS_CODES + i;
    else
    {
        /* Register the class */
        if(nClass == MAX_CLASSES) {LexErrorCode = LEXERR_TOO_MANY_CLASSES; return FALSE;}
        for(i = 0; i < MAX_CHARS; i++) hpClass[nClass][i] = Tab[i];
        *ClassCode = CLASS_CODES + nClass;
        nClass++;
    }

    *Clength = Index;
    return TRUE;
}

/* Expand the regular expression at position hpExpr, replacing the IDs between { and } by
   their definition (processing Part 1 or 2). Write the result to array lpResult. New position in the
   array is returned via NewPos. New position in the hpExpr buffer is returned via NewPosExpr.
   NOTE: Expansion of expressions of Part 1 is done immediately as Part 1 is processed, so there is no
   need for recursive expansion calls, since an identifier appearing in an expression must have been
   declared in a line above. Classes are replaced by class codes (class number + CLASS_CODES), the class
   number being an index to hpClass, which describes the classes. */
BOOL ExpandExpression(LPHUGE hpExpr, LPWORD lpResult, DWORD LimitPos, DWORD *NewPos, DWORD *NewPosExpr)
{
  register DWORD i,j,iResult;
  DWORD i2;
  static WORD Length,ClassCode;
  unsigned char c;
  BOOL AddCaseChar;

  i = 0; iResult = 0;
  while(1)
  {
    switch(hpExpr[i])
    {
      case '\\':
                // Backslash single character quotation not in string, not in class: may escape as in C
                c = BackslashConstant(hpExpr + i,(LPWORD)&Length);
                if(CaseSensitive == FALSE)
                {
                    AddCaseChar = FALSE;
                    if(c >= 'A' && c <= 'Z') {c += 'a' - 'A'; AddCaseChar = TRUE;}
                    else
                        if(c >= 'a' && c <= 'z') {c -= 'a' - 'A'; AddCaseChar = TRUE;}
                    if(AddCaseChar) Length += 4;    // Add surrounding (), | and a letter
                }
                if(iResult + Length > LimitPos)
                {
                    LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
                    LexErrorIntParam = MAX_REG_EXP;
                    return FALSE;
                }
                if(CaseSensitive == FALSE && AddCaseChar) lpResult[iResult++] = '(';
                for(j = 0; j < Length; j++) lpResult[iResult++] = hpExpr[i + j];
                if(CaseSensitive == FALSE && AddCaseChar)
                {
                    lpResult[iResult++] = '|';
                    lpResult[iResult++] = 'c';
                    lpResult[iResult++] = ')';
                }
                i += Length - 1; yycol += Length - 1;
                break;
      case '.':
      case '[': // A character range!
                if(ExpandClass(hpExpr + i, (LPWORD)&ClassCode, (LPWORD)&Length) == FALSE) return FALSE;
                i += Length;
                if(iResult > LimitPos)
                {
                    LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
                    LexErrorIntParam = MAX_REG_EXP;
                    return FALSE;
                }
                lpResult[iResult] = ClassCode;
                iResult++;
                yycol += Length;
                break;
      case '"': /* Copy characters up to matching ". No case insensitivity inside strings! Beware of the sequence \" */
                LexErrorLine = yyline;
                LexErrorColumn = yycol;
                LexErrorLineString = yyline;
                LexErrorColString = yycol;
                if(iResult > LimitPos)
                {
                    LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
                    LexErrorIntParam = MAX_REG_EXP;
                    return FALSE;
                }
                lpResult[iResult] = hpExpr[i];
                iResult++;
                i++; yycol++;
                CONT_SCAN:
                while((c = hpExpr[i]) && c != 10 && c != '\\' && c != '"')
                {
                    if(iResult > LimitPos)
                    {
                        LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
                        LexErrorIntParam = MAX_REG_EXP;
                        return FALSE;
                    }
                    lpResult[iResult++] = hpExpr[i];
                    if(hpExpr[i] != 13) yycol++;
                    i++;
                }
                if(hpExpr[i] == 10)
                {
                    LexErrorCode = LEXERR_REACHED_EOL_STRING;
                    return FALSE;
                }
                if(hpExpr[i] == 0)
                {
                    LexErrorCode = LEXERR_REACHED_EOF_STRING;
                    return FALSE;
                }
                if(hpExpr[i] == '\\')
                {
                    if(hpExpr[i+1] == 0) {LexErrorCode = LEXERR_REACHED_EOF_BSLASH; return FALSE;}
                    if(iResult + 1 > LimitPos)
                    {
                        LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
                        LexErrorIntParam = MAX_REG_EXP;
                        return FALSE;
                    }
                    lpResult[iResult] = '\\';
                    i++; yycol++; iResult++;
                    lpResult[iResult] = hpExpr[i];
                    i++; yycol++; iResult++; goto CONT_SCAN;
                }

                /* Last matching " */
                if(iResult > LimitPos)
                {
                    LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
                    LexErrorIntParam = MAX_REG_EXP;
                    return FALSE;
                }
                lpResult[iResult] = hpExpr[i];
                iResult++;
                break;
      case 10:  yyline++; yycol = 1;
      case 13:
      case 0:
      case ' ':
      case '\t':*NewPos = iResult;
                *NewPosExpr = i;
                return TRUE;
      case '{': /* Operate an ID substitution! No case insensitivity while doing ID substitution! */
                LexErrorLine = yyline;
                yycol++;
                LexErrorColumn = yycol;
                i2 = i; if(ReadId(hpExpr + i + 1, NewPosExpr) == FALSE) return FALSE;
                i2 += *NewPosExpr + 1;
                if(hpExpr[i2] == 0)     {LexErrorCode = LEXERR_REACHED_EOF_ID; return FALSE;}
                if(hpExpr[i2] != '}')   {LexErrorCode = LEXERR_MISSING_RIGHT_BRACE; return FALSE;}
                if(ReplaceId(hpExpr + i + 1, (WORD)(i2 - i - 1), lpResult + iResult, LimitPos-iResult, NewPos) == FALSE) return FALSE;
                iResult += *NewPos;
                i = i2;
                break;
      default:  /* Simple character copy. Case sensitivity matters! */
                if(CaseSensitive == FALSE)
                {
                    AddCaseChar = FALSE;
                    c = hpExpr[i];
                    if(c >= 'A' && c <= 'Z') {c += 'a' - 'A'; AddCaseChar = TRUE;}
                    else
                        if(c >= 'a' && c <= 'z') {c -= 'a' - 'A'; AddCaseChar = TRUE;}
                    if(AddCaseChar) Length = 4; // Because (x|X)
                }
                else Length = 0;

                if(iResult + Length > LimitPos)
                {
                    LexErrorCode = LEXERR_REG_EXP_TOO_LARGE;
                    LexErrorIntParam = MAX_REG_EXP;
                    return FALSE;
                }
                if(CaseSensitive == FALSE && AddCaseChar) lpResult[iResult++] = '(';
                lpResult[iResult++] = hpExpr[i];
                if(CaseSensitive == FALSE && AddCaseChar)
                {
                    lpResult[iResult++] = '|';
                    lpResult[iResult++] = c;
                    lpResult[iResult++] = ')';
                }
                break;
    }
    if(hpExpr[i] != 13) yycol++; i++;
  }
}

/* Prepare the data structures with data from input buffer. */
BOOL LexPrepare()
{
  static DWORD Index,Index2,Index3;
  WORD base,level,j;
  BOOL AfterWhiteSpace;

  Index = Index2 = Index3 = 0;

  /* Part 1: identifier declarations as regular expressions, and zero or one code fragment enclosed in { }.
     The code fragment is copied to the produced file, without the enclosing { }. */
  while(1)
  {
    AfterWhiteSpace = FALSE;
    switch(hpInputBuffer[Index])
    {
      case 0:   LexErrorCode = LEXERR_NO_RULE_IN_BUFFER; return FALSE;
      case '%': Index++; yycol++;
                if(hpInputBuffer[Index] == '%') {Index++; yycol++; goto PART2;}
                else {LexErrorCode = LEXERR_INVALID_CHARACTER; LexErrorCharParam = hpInputBuffer[Index]; return FALSE;}
      case '{': P1Code.hpCodeStart = hpInputBuffer + Index;
                P1Code.LineNumber = yyline;
                if(ReadCode(hpInputBuffer + Index,&Index2) == FALSE) return FALSE;
                P1Code.CodeSize = Index2 + 1;
                Index += Index2;
                break;
      case ' ':
      case 13:
      case '\t':
      case 10:  if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                Index += Index2 - 1;
                AfterWhiteSpace = TRUE;
                break;

      default:  /*** A declaration or a comment ***/
                if(hpInputBuffer[Index] == '/' && (hpInputBuffer[Index+1] == '*' || hpInputBuffer[Index+1] == '/'))
                {
                    /* A Comment ! */
                    if(ReadComment(hpInputBuffer+Index,&Index2) == FALSE) return FALSE;
                    Index += Index2;
                    break;
                }

                /* A declaration */
                if(AddP1Declaration() == FALSE) return FALSE;

                /* Read and register identifier */
                lpP1Decl[nP1Decl-1].lp1IdStart = hpInputBuffer + Index;
                if(ReadId(hpInputBuffer + Index, &Index2) == FALSE) return FALSE;
                lpP1Decl[nP1Decl-1].IdSize = (WORD)Index2;
                Index += Index2;
                if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE; // Skip Whitespace
                if(Index2 == 0)
                {
                    LexErrorCode = LEXERR_MISSING_WHITESPACE;
                    return FALSE;
                }
                Index += Index2;

                /* Read, expand and register regular expression */
                nP1Decl--;  // Temporarily mask the current declaration, so it can't be seen during the expansion process
                if(ExpandExpression(
                    hpInputBuffer+Index,
                    (LPWORD)(lpP1Decl[nP1Decl].RegExp),
                    MAX_REG_EXP-1,
                    &Index2,
                    &Index3) == FALSE) return FALSE;
                lpP1Decl[nP1Decl].RegExp[Index2] = 0;   // Add trailing zero !
                Index += Index3;
                lpP1Decl[nP1Decl].RegExpSize = (WORD)Index2;
                nP1Decl++;  // Restore the current declaration
    }
    if(AfterWhiteSpace == FALSE) yycol++;
    Index++;
  }

  /* Part 2: List of token definitions as regular expressions, each associated to an action (portion of C code).
     Action will later by copied each into the case of a switch statement, and the enclosing { } will be removed.
     NOTE: as most actions are simply to return a token value, one can replace the code fragment:
     
            {return TOKEN_VALUE;}
     
     by:
            TOKEN_VALUE
            
     alone. This way, the switch statement will not be overloaded with dozens of return statements.
  */
  PART2:
  while(1)
  {
    AfterWhiteSpace = FALSE;
    switch(hpInputBuffer[Index])
    {
        case 0:     if(nP2RegExp == 0) {LexErrorCode = LEXERR_NO_RULE_IN_BUFFER; return FALSE;}
                    else return TRUE;   // No Part 3
        case ' ':
        case 13:
        case '\t':
        case 10:    if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE; // Skip Whitespace
                    Index += Index2 - 1;
                    AfterWhiteSpace = TRUE;
                    break;
        case '%':   Index++; yycol++;
                    if(hpInputBuffer[Index] == 'I') // Turn case sensitivity to FALSE
                    {
                        CaseSensitive = FALSE;
                        break;
                    }
                    if(hpInputBuffer[Index] == 'S') // Turn case sensitivity to TRUE (default)
                    {
                        CaseSensitive = TRUE;
                        break;
                    }
                    if(hpInputBuffer[Index] == '%')
                    {
                        Index++;
                        yycol++;
                        if(nP2RegExp == 0) {LexErrorCode = LEXERR_NO_RULE_IN_BUFFER; return FALSE;}
                        goto PART3;
                    }
                    else {LexErrorCode = LEXERR_INVALID_CHARACTER; LexErrorCharParam = hpInputBuffer[Index]; return FALSE;}

        default:    /*** A token definition or a comment ***/
                    if(hpInputBuffer[Index] == '/' && (hpInputBuffer[Index+1] == '*' || hpInputBuffer[Index+1] == '/'))
                    {
                        /* A Comment ! */
                        if(ReadComment(hpInputBuffer+Index,&Index2) == FALSE) return FALSE;
                        Index += Index2;
                        break;
                    }

                    /* A token definition ! */
                    if(AddP2RegularExpression() == FALSE) return FALSE;

                    /* Read, expand and register regular expression. Add parenthesis and an endmarker '#'! */
                    hpP2RegExp[nP2RegExp-1].RegExpLineNumber = yyline;
                    hpP2RegExp[nP2RegExp-1].RegExpColumnNumber = yycol;

                    hpP2RegExp[nP2RegExp-1].RegExp[0] = '(';
                    if(ExpandExpression(
                        hpInputBuffer+Index,
                        (LPWORD)(hpP2RegExp[nP2RegExp-1].RegExp + 1),   // + 1 to skip the left parenthesis if exists
                        MAX_REG_EXP - 4,    // Because of the 2 parenthesis, the endmarker and the trailing zero !!
                        &Index2,
                        &Index3) == FALSE) return FALSE;
                    hpP2RegExp[nP2RegExp-1].RegExp[Index2+1] = ')';
                    hpP2RegExp[nP2RegExp-1].RegExp[Index2+2] = '#'; // Add endmarker
                    hpP2RegExp[nP2RegExp-1].RegExp[Index2+3] = 0;   // Add trailing zero !
                    Index += Index3;
                    hpP2RegExp[nP2RegExp-1].RegExpSize = (WORD)Index2 + 3;

                    base = 1;
                    level = 0;
                    for(j = 1; j < hpP2RegExp[nP2RegExp-1].RegExpSize - 2; j++)
                    {
                        if(hpP2RegExp[nP2RegExp-1].RegExp[j] == '(') level++;
                        else
                            if(hpP2RegExp[nP2RegExp-1].RegExp[j] == ')')
                            {
                                if(j == hpP2RegExp[nP2RegExp-1].RegExpSize - 3)
                                    base = 0;                   // No parenthesis if surrounding parenthesis already
                                else
                                {
                                    level--;
                                    if(level == 0) break;       // Mandatory surrounding parenthesis!
                                }
                            }
                    }

                    if(base == 0)
                    {
                        /* Eliminate redundant surrounding parenthesis */
                        for(j = 0; j < hpP2RegExp[nP2RegExp-1].RegExpSize - 3; j++)
                            hpP2RegExp[nP2RegExp-1].RegExp[j] = hpP2RegExp[nP2RegExp-1].RegExp[j + 1];
                        hpP2RegExp[nP2RegExp-1].RegExp[hpP2RegExp[nP2RegExp-1].RegExpSize - 3] = '#';
                        hpP2RegExp[nP2RegExp-1].RegExp[hpP2RegExp[nP2RegExp-1].RegExpSize - 2] = 0;
                        hpP2RegExp[nP2RegExp-1].RegExpSize -= 2;
                    }

                    /* Read and register action and line number */
                    if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE; // Skip Whitespace
                    if(Index2 == 0) {LexErrorCode = LEXERR_MISSING_WHITESPACE; return FALSE;}
                    Index += Index2;
                    hpP2RegExp[nP2RegExp-1].Action.LineNumber = yyline;
                    hpP2RegExp[nP2RegExp-1].Action.hpCodeStart = hpInputBuffer + Index;
                    if(hpInputBuffer[Index] == '{')
                    {
                        /* Read a code fragment */
                        if(ReadCode(hpInputBuffer + Index, &Index2) == FALSE) return FALSE;
                        hpP2RegExp[nP2RegExp-1].Action.CodeSize = Index2 + 1;
                    }
                    else
                    {
                        /* Read a token to return */
                        if(ReadId(hpInputBuffer + Index, &Index2) == FALSE) return FALSE;
                        hpP2RegExp[nP2RegExp-1].Action.CodeSize = -(long int)Index2;  // Negative code size means return token
                    }
                    Index += Index2;

                    break;
    }
    if(AfterWhiteSpace == FALSE) yycol++;
    Index++;
  }

  /* Part 3: support routines and code. Need not be enclosed in { }. */
  PART3:
  if(hpInputBuffer[Index] == 0) return TRUE;    // No code in Part 3
  P3Code.LineNumber = yyline;
  P3Code.hpCodeStart = hpInputBuffer + Index;
  Index2 = Index;
  while(hpInputBuffer[Index2]) Index2++;
  P3Code.CodeSize = Index2 - Index;

  return TRUE;
}
