/******************************************************************
* Phoenix Lexical Analyzer Generator.                             *
* Minimizing the global DFA.                                      *
******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "ph_scan.h"
#include "ph_error.h"

static WORD nGroups;

/* Create initial partition:
   1. Gather accepting states recognizing the same token into small groups (some tokens may be recognized by
      several states, as in regular expression A*).
   2. Put all other states together into a big group.
   3. Determine number of groups. */
void CreateInitialPartition()
{
    register WORD i;

    nGroups = 0;
    for(i = 1; i <= nStates; i++)   // Loop on all states
    {
        hpFNextState[i].Group = lpStates[i].Token;                      // Points 1. and 2. (used recognized token!)
        if(lpStates[i].Token > nGroups) nGroups = lpStates[i].Token;    // Point 3.
    }
}

/* Compute final minimized transition table hpGNextState, let it have TotalStates entries. */
BOOL Transitions()
{
    register WORD i,j;
    WORD l,representative;

    TotalStates = nGroups;  // New valid states indexed from 1 to nGroups included

    /* Clear hpGNextState */
    for(i = 0; i <= TotalStates; i++)
    {
        hpGNextState[i].Token = 0;
        for(j = 0; j < MAX_CHARS; j++) hpGNextState[i].NextState[j] = 0;
        for(j = 0; j < MAX_CLASSES; j++) hpGNextState[i].NextStateClass[j] = 0;
    }

    for(i = 1; i <= TotalStates; i++)   // Loop on new states
    {
        for(j = 1; j <= nStates; j++) if(hpFNextState[j].Group == i) break; // Find a representative for the group
        representative = j;

        for(j = 1; j < MAX_CHARS; j++)  // Loop on characters
        {
            l = hpFNextState[representative].NextState[j];
            hpGNextState[i].NextState[j] = hpFNextState[l].Group;
        }

        for(j = 0; j < nClass; j++)     // Loop on classes
        {
            l = hpFNextState[representative].NextStateClass[j];
            hpGNextState[i].NextStateClass[j] = hpFNextState[l].Group;
        }
        hpGNextState[i].Token = lpStates[representative].Token;
    }

    return TRUE;
}

/* Compute final partitionment */
BOOL Pinew()
{
    register WORD i,j,k;
    BOOL SameGroup,BreakAGroup;
    WORD l,m,BreakGroup;
    LPWORD lpWorkGroup;
    WORD nWorkGroup;

    /* Allocate memory for workgroup */
    lpWorkGroup = (LPWORD)GlobalAllocPtr(GHND,(DWORD)sizeof(WORD) * (DWORD)nStates);
    if(lpWorkGroup == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    /* The dead state is state zero, which has transitions to itself on all
       input symbols. */
    lpStates[0].Subset = 0;

    do
    {
        BreakAGroup = FALSE;
        for(i = 0; i <= nGroups; i++)           // Loop on groups in the partition, starting at last group nGroups
        {
TRY_AGAIN_SAME_GROUP:
            nWorkGroup = 0;                     // Set workgroup to empty

            for(j = 1; j <= nStates; j++)       // Put group together
                if(hpFNextState[j].Group == i)
                {nWorkGroup++; lpWorkGroup[nWorkGroup - 1] = j;}

            if(nWorkGroup == 0) continue;

            if(nWorkGroup == 1)
            {
                if(i == 0)      // Break group 0, and replace it with a new group
                {
                    nGroups++;
                    hpFNextState[lpWorkGroup[0]].Group = nGroups;
                }
                continue;       // If group has only 1 element, it is not breakable
            }

            /* Loop on states of group from the second to the last, comparing each state to the first one.
               If, for all characters, the two states compared have transitions to the same groups,
               then put them into a new group. This way, the first state will break the initial group,
               and attract to it's new group some other states. */
            BreakGroup = 0; // Number of states to leave their initial group number i
            nGroups++;
            hpFNextState[lpWorkGroup[0]].Group = nGroups;   // Suppose first state leaves the group
            for(j = 1; j < nWorkGroup; j++)
            {

                /* Now comparing lpWorkGroup[0] to lpWorkGroup[j] */

                SameGroup = TRUE;
                for(k = 1; k < MAX_CHARS; k++)  // Loop on characters
                {
                    /* Transition for first state of group, on character k */
                    l = hpFNextState[lpWorkGroup[0]].NextState[k];

                    /* Transition for state j of group, on character k */
                    m = hpFNextState[lpWorkGroup[j]].NextState[k];

                    if(hpFNextState[l].Group != hpFNextState[m].Group)  // Do transitions go to the same group ?
                    {SameGroup = FALSE; break;}
                }

                if(SameGroup == FALSE) continue;                        // States just compared are NOT in the same group

                for(k = 1; k < nClass; k++) // Loop on classes
                {
                    /* Transition for first state of group, on class k */
                    l = hpFNextState[lpWorkGroup[0]].NextStateClass[k];

                    /* Transition for state j of group, on character k */
                    m = hpFNextState[lpWorkGroup[j]].NextStateClass[k];

                    if(hpFNextState[l].Group != hpFNextState[m].Group)  // Do transitions go to the same group ?
                    {SameGroup = FALSE; break;}
                }

                if(SameGroup == FALSE) continue;                        // States just compared are NOT in the same group

                /*** States just compared are in the same group! ***/

                /* Is the state lpWorkGroup[j] the last one leaving the initial group ? */
                if(i != 0 && BreakGroup == nWorkGroup - 2)  // Last state leaving the initial group
                {
                    /* Take the initial group's number, instead of the new group's number, in order to
                       avoid creating a hole in the numbers (but i != 0, so group zero must exceptionally
                       be left empty!). */
                    for(k = 0; k < nWorkGroup; k++) hpFNextState[lpWorkGroup[k]].Group = i;
                    nGroups--;  // Delete the previously created new group number
                }
                else
                {
                    /* Move the state lpWorkGroup[j] to the new group */
                    hpFNextState[lpWorkGroup[j]].Group = hpFNextState[lpWorkGroup[0]].Group;
                }
                BreakGroup++;
            }
            if(BreakGroup == 0)
            {
                /* The first state of the group does not belong to the same group than all other
                   states of the group: put it in a separate one. */
                BreakAGroup = TRUE;
            }
            else
                if(BreakGroup < nWorkGroup - 1)         // If not complete group was broken (some states remain),
                {
                    BreakAGroup = TRUE;
                    goto TRY_AGAIN_SAME_GROUP;          // try to break rest of group i immediately
                }
        }
    }
    while(BreakAGroup);                         // Exit do loop when no more group has been broken

    /* Free memory of work group */
    GlobalFreePtr(lpWorkGroup);

    return TRUE;
}

BOOL MinimizeDFA()
{
    if(Pinew() == FALSE) return FALSE;
    if(Transitions() == FALSE) return FALSE;
    DfaStart = hpFNextState[1].Group;

    return TRUE;
}
