/******************************************************************
* Phoenix Parser Generator.                                       *
******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "ph_scan.h"        // For input buffer declarations
#include "ph_parse.h"
#include "ph_error.h"

extern HINSTANCE hDLLInst;                  // The handle of the DLL instance
static CLOSED_SET_OF_ITEMS CIset;           // Temporary set of items
static SET_OF_ITEMS Iset;                   // Temporary set of items
static HFILE hf;                            // The state file giving the trace of actions and items
static char StateFile[MAX_FILE_NAME + 1];   // Complete path of state file
static DWORD NewHead;


/* Parser generator Initializations */
BOOL PInit()
{
    hpInputBuffer   = NULL;
    hpSymbol        = NULL;
    hpProduction    = NULL;
    hpCsiTable      = NULL;
    hpActionGoto    = NULL;
    hpPropa         = NULL;
    ActionState     = NULL;
    GotoSymbol      = NULL;
    AGPool          = NULL;
    lpPool          = NULL;
    Holes           = 0;
    nTerm           = 0;
    nSymbol         = 0;
    nMaxSymbol      = 0;
    nProduction     = 0;
    nMaxProduction  = 0;
    nCsi            = 0;
    nMaxCsi         = 0;
    nPropa          = 0;
    nMaxPropa       = 0;
    nActionGoto     = 0;
    nMaxActionGoto  = 0;
    yyline          = 1;
    yycol           = 1;

    return TRUE;
}

/* Memory management to do before exiting program */
BOOL PEnd()
{
    if(hpInputBuffer != NULL)
    {
        GlobalFreePtr(hpInputBuffer);
        hpInputBuffer = NULL;
    }

    if(hpSymbol != NULL)
    {
        GlobalFreePtr(hpSymbol);
        hpSymbol = NULL;
    }

    if(hpProduction != NULL)
    {
        GlobalFreePtr(hpProduction);
        hpProduction = NULL;
    }

    if(hpCsiTable != NULL)
    {
        GlobalFreePtr(hpCsiTable);
        hpCsiTable = NULL;
    }

    if(hpActionGoto != NULL)
    {
        GlobalFreePtr(hpActionGoto);
        hpActionGoto = NULL;
    }

    if(hpPropa != NULL)
    {
        GlobalFreePtr(hpPropa);
        hpPropa = NULL;
    }

    if(lpPool != NULL)
    {
        GlobalFreePtr(lpPool);
        lpPool = NULL;
    }

    FreePackedArrays();

    return TRUE;
}

/* Compute, for each nonterminal C, the list of indexes to nonterminals A such as C => As with a rightmost derivation */
BOOL ComputeDescendants()
{
    register WORD i,j,k,m;
    WORD iSymbol,Left,Descendant;

    for(i = 0; i < nProduction; i++)            // Loop on productions
    {
        /* The production must have at least one nonterminal on the right side */
        if(hpProduction[i].Length && (iSymbol = hpProduction[i].Right[0]) >= nTerm)
        {
            Left = hpProduction[i].Left;
            for(j = 0; j < hpSymbol[Left].nDesc; j++) if(hpSymbol[Left].Descendants[j] == iSymbol) break;
            if(j < hpSymbol[Left].nDesc) continue;      // Descendant already in set!
            if(hpSymbol[Left].nDesc >= NDESC)
            {
                LexErrorCode = PERR_NDESC_TOO_LOW;
                return FALSE;
            }
            hpSymbol[Left].Descendants[hpSymbol[Left].nDesc++] = iSymbol;
        }
    }

    /* Transitivity algorithm */
    for(i = nTerm + 1; i < nSymbol; i++)        // Loop on nonterminals (do not compute descendants of S' !)
    {
        /* Each nonterminal is a descendant of itself! */
        for(j = 0; j < hpSymbol[i].nDesc; j++) if(hpSymbol[i].Descendants[j] == i) break;   // already in set of descendants
        if(j >= hpSymbol[i].nDesc)
        {
            if(hpSymbol[i].nDesc >= NDESC)
            {
                LexErrorCode = PERR_NDESC_TOO_LOW;
                return FALSE;
            }
            hpSymbol[i].Descendants[hpSymbol[i].nDesc++] = i;
        }

        /* Compute descendants by transitivity */
        for(j = 0; j < hpSymbol[i].nDesc; j++)  // Loop on descendants
        {
            Descendant = hpSymbol[i].Descendants[j];
            if(Descendant == i) continue;       // Do not consider descendants of the nonterminal with direct recursion!

            /* Add the descendants of the current descendant */
            for(k = 0; k < hpSymbol[Descendant].nDesc; k++)
            {
                for(m = 0; m < hpSymbol[i].nDesc; m++)
                    if(hpSymbol[i].Descendants[m] == hpSymbol[Descendant].Descendants[k]) break;    // Already in set!
                if(m >= hpSymbol[i].nDesc)
                {
                    if(hpSymbol[i].nDesc >= NDESC)
                    {
                        LexErrorCode = PERR_NDESC_TOO_LOW;
                        return FALSE;
                    }
                    hpSymbol[i].Descendants[hpSymbol[i].nDesc++] = hpSymbol[Descendant].Descendants[k];
                }
            }
        }
    }

    return TRUE;
}

/* Function First(X), called at the begining for initialization:
   First(X) is computed for every X in the grammar. May also be
   called later to compute First(string of symbols). */
BOOL First(
    long int HOST_FAR *X,   // Array of indexes to hpSymbol: 1 or n symbols. May be < 0, because of P_EMPTY!
    WORD Length,            // Number of symbols in X (1 for First(symbol), n for First(string))
    BOOL Init,              // TRUE if initialization, FALSE to compute First(string)
    long int HOST_FAR *FX,  // Output: array of symbols (indexes to hpSymbol). May be < 0, because of P_EMPTY!
    LPWORD nFX              // Output: number of entries in FX
    )
{
    register WORD i,j,k,m;
    BOOL Added,FoundEmpty;
    LPPRODUCTION hpProd;
    LPSYMBOL hpG;
    WORD iSymb,iTerm;

    if(Init)    // Compute First(X) for X = grammar symbol
    {
        for(i = 0; i < nTerm; i++)
        {hpSymbol[i].nTerm = 1; hpSymbol[i].First[0] = i;}      // Reset First(X) for every X = terminal
        for(i = nTerm; i < nSymbol; i++) hpSymbol[i].nTerm = 0; // Reset First(X) for every X = nonterminal

        do
        {
            Added = FALSE;

            /* Add to First(X) according to productions G -> X1 X2 ... Xn */
            for(i = 0; i < nProduction; i++)            // Loop on productions
            {
                hpProd = hpProduction + i;              // Points to current production
                hpG = hpSymbol + hpProd -> Left;        // Points to current nonterminal G (left side of prod), in hpSymbol
                for(j = 0; j < hpProd -> Length; j++)   // Loop on symbols Xj to the right of the production
                {
                    iSymb = hpProd -> Right[j];

                    /*** Add First(Xj) to First(G) ***/
                    FoundEmpty = FALSE;
                    for(k = 0; k < hpSymbol[iSymb].nTerm; k++)  // Loop on terminals in First(Xj)
                    {
                        iTerm = hpSymbol[iSymb].First[k];
                        if(iTerm == P_EMPTY)
                            FoundEmpty = TRUE;                  // P_EMPTY is in First(Xj)
                        else
                        {
                            /*** Add terminal to First(G) if not already in it ***/
                            for(m = 0; m < hpG -> nTerm; m++) if(hpG -> First[m] == (short int)iTerm) break;
                            if(m >= hpG -> nTerm)
                            {
                                if(hpG -> nTerm == NTERM) {LexErrorCode = PERR_NTERM_TOO_SHORT; return FALSE;}
                                hpG -> First[hpG -> nTerm++] = iTerm;
                                Added = TRUE;
                            }
                        }
                    }
                    if(FoundEmpty == FALSE) break;      // If P_EMPTY is not in First(Xj), then terminals in First(Xj+k)
                                                        // with k > 0, cannot be in First(G)
                }
                if(FoundEmpty || hpProd -> Length == 0) // If First(Xn) contains P_EMPTY or the right side is empty,
                                                        // add P_EMPTY to First(G)
                {
                    for(m = 0; m < hpG -> nTerm; m++) if(hpG -> First[m] == P_EMPTY) break;
                    if(m >= hpG -> nTerm)
                    {
                        if(hpG -> nTerm == NTERM) {LexErrorCode = PERR_NTERM_TOO_SHORT; return FALSE;}
                        hpG -> First[hpG -> nTerm++] = P_EMPTY;
                        Added = TRUE;
                    }
                }
            }
        } while(Added);
    }
    else        // Compute First(X) for X = string of symbols
    {   
        *nFX = 0;
        for(i = 0; i < Length; i++)                     // Loop on symbols Xi in the string
        {
            FoundEmpty = FALSE;
            if(X[i] == P_EMPTY) FoundEmpty = TRUE;
            else
            {
                for(j = 0; j < hpSymbol[X[i]].nTerm; j++)   // Loop on terminals in First(Xi)
                {
                    iTerm = hpSymbol[X[i]].First[j];
                    if(iTerm == P_EMPTY)
                        FoundEmpty = TRUE;                  // P_EMPTY is in First(Xi)
                    else
                    {
                        /*** Add terminal to First(string) if not already in it ***/
                        for(m = 0; m < *nFX; m++) if(FX[m] == (short int)iTerm) break;
                        if(m >= *nFX)
                        {
                            if(*nFX == NTERM) {LexErrorCode = PERR_NTERM_TOO_SHORT; return FALSE;}
                            FX[(*nFX)++] = iTerm;
                        }
                    }
                }
            }
            if(FoundEmpty == FALSE) break;  // If P_EMPTY is not in First(Xi), then terminals in First(Xi+k)
                                            // with k > 0, cannot be in First(string)
        }

        if(FoundEmpty || Length == 0)   // If First(Xn) contains P_EMPTY or if the string is empty, add P_EMPTY to First(string)
        {
            for(m = 0; m < *nFX; m++) if(FX[m] == P_EMPTY) break;
            if(m >= *nFX)
            {
                if(*nFX == NTERM) {LexErrorCode = PERR_NTERM_TOO_SHORT; return FALSE;}
                FX[(*nFX)++] = P_EMPTY;
            }
        }
    }

    return TRUE;    
}

/* Add an item to a closed set */
BOOL AddItemClosed(
    WORD Prod,          // Index of the production
    WORD Pos,           // Position of the dot in the production
    LPCSITEMS hpSi      // Pointer to a closed set of items
    )
{
    register WORD i;
    LPLR1ILIST hpLr;

    /* Find the index in the list of LR1_ITEM_LIST structures of the one that has for production Prod
       and for position Pos */
    for(i = 0; i < hpSi -> Number; i++)
    {
        hpLr = hpSi -> List + i;
        if(hpLr -> Prod == Prod && hpLr -> Pos == Pos) break;
    }

    if(i >= hpSi -> Number)
    {
        /* New couple (production, position) */
        if(hpSi -> Number == MAX_ITEMS)
        {
            LexErrorCode = PERR_MAX_ITEMS_TOO_SHORT;
            return FALSE;
        }
        hpSi -> List[hpSi -> Number].Prod = Prod;
        hpSi -> List[hpSi -> Number].Pos = Pos;
        hpSi -> List[hpSi -> Number].Term = 0;
        hpSi -> List[hpSi -> Number].nPropa = 0;
        hpSi -> Number++;
    }

    return TRUE;    
}

/* Add an item to a set */
BOOL AddItem(
    WORD Prod,          // Index of the production
    WORD Pos,           // Position of the dot in the production
    LPSITEMS hpSi       // Pointer to a set of items
    )
{
    register WORD i;
    LPLR1ILIST hpLr;

    /* Find the index in the list of LR1_ITEM_LIST structures of the one that has for production Prod
       and for position Pos */
    for(i = 0; i < hpSi -> Number; i++)
    {
        hpLr = hpSi -> List + i;
        if(hpLr -> Prod == Prod && hpLr -> Pos == Pos) break;
    }

    if(i >= hpSi -> Number)
    {
        /* New couple (production, position) */
        if(hpSi -> Number == COUPLES)
        {
            LexErrorCode = PERR_COUPLES_TOO_SHORT;
            return FALSE;
        }
        hpSi -> List[hpSi -> Number].Prod = Prod;
        hpSi -> List[hpSi -> Number].Pos = Pos;
        hpSi -> List[hpSi -> Number].Term = 0;
        hpSi -> List[hpSi -> Number].nPropa = 0;
        hpSi -> Number++;
    }

    return TRUE;    
}

/* The following routine returns the item number if the item Prod,Pos is in the closed set of items hpSi,
   else returns -1 */
int ItemIsInClosed(
    WORD Prod,          // Index of the production
    WORD Pos,           // Position of the dot in the production
    LPCSITEMS hpSi      // Pointer to a closed set of items
    )
{
    register WORD i;
    LPLR1ILIST hpLr;

    for(i = 0; i < hpSi -> Number; i++)     // Loop on items of hpSi
    {
        hpLr = hpSi -> List + i;
        if(hpLr -> Prod == Prod && hpLr -> Pos == Pos) return i;
    }

    return -1;
}

/* The following routine returns the item number if the item Prod,Pos is in the set of items hpSi,
   else returns -1 */
int ItemIsIn(
    WORD Prod,          // Index of the production
    WORD Pos,           // Position of the dot in the production
    LPSITEMS hpSi       // Pointer to a set of items
    )
{
    register WORD i;
    LPLR1ILIST hpLr;

    for(i = 0; i < hpSi -> Number; i++)     // Loop on items of hpSi
    {
        hpLr = hpSi -> List + i;
        if(hpLr -> Prod == Prod && hpLr -> Pos == Pos) return i;
    }

    return -1;
}

static WORD ly;
static WORD lfy;
static long int HOST_FAR *lpy;
static long int HOST_FAR *lpfy;

/* Closure function. Computes the closure of a set of items limited to one item. The purpose is to discover
   propagations and spontaneoulsy generated lookaheads, so the input item is given without lookahead. */
BOOL Closure(
    DWORD Set,      // Index of the set containing the item
    DWORD Item,     // Index of the item in the set
    LPCSITEMS hpS   // The resulting set (closure)
    )
{
    register DWORD i,j,k,m,lk;
    LPLR1ILIST hpLr;
    DWORD Prod,Pos;
    WORD iSymb,itm;

    hpLr = (hpCsiTable + Set) -> List + Item;

    /* Initialize the resulting set with the item itself */
    hpS -> Number = 1;
    hpS -> List[0].Prod = hpLr -> Prod;
    hpS -> List[0].Pos  = hpLr -> Pos;
    hpS -> List[0].Propa = 0;
    hpS -> List[0].nPropa = 0;
    hpS -> List[0].Term = 0;
    if(PoolAddPosToList(hpS -> List[0].Term, P_EMPTY, (LPDWORD)&NewHead) == FALSE) return FALSE;
    hpS -> List[0].Term = NewHead;

    /* Loop on the items of the set. Initially, the set has 1 item, but other items are added during the process */
    for(i = 0; i < hpS -> Number; i++)  // Item [A -> a.Br,c]
    {
        hpLr = hpS -> List + i;
        Prod = hpLr -> Prod;
        Pos  = hpLr -> Pos;

        /* There must be a nonterminal B after the dot, to explore subproductions */
        iSymb = hpProduction[Prod].Right[Pos];
        if(iSymb < nTerm) continue;         // Closure limited to the item itself, because no nonterminal after the dot!

        lk = hpLr -> Term;
        while(lk)                           // Loop on the lookaheads, to compose the different LR(1) items
        {
            /* Compute First(rc) */
            ly = 0;
            for(j = Pos + 1; j < hpProduction[Prod].Length; j++) lpy[ly++] = hpProduction[Prod].Right[j];   // Build r
            lpy[ly++] = lpPool[lk].Pos;
            if(First(lpy,ly,FALSE,lpfy,(LPWORD)&lfy) == FALSE) return FALSE;

            /* Find productions having B as left side: B -> v, using the special index LeftProd of the symbol table */
            for(j = 0; j < hpSymbol[iSymb].nLeftProd; j++)  // Loop on all productions having B as left side
            {
                if((itm = ItemIsInClosed(hpSymbol[iSymb].LeftProd[j],0,hpS)) == -1) // The item [B -> .v] is not in the set
                {
                    if(AddItemClosed(hpSymbol[iSymb].LeftProd[j],0,hpS) == FALSE) return FALSE;
                    itm = hpS -> Number - 1;
                }

                /* Take the terminals b of First(rc) */
                for(k = 0; k < lfy; k++)
                {
                    m = hpS -> List[itm].Term;
                    while(m)
                    {
                        if(lpPool[m].Pos == lpfy[k]) break; // Terminal already present
                        m = lpPool[m].Next;
                    }

                    if(m == 0)
                    {
                        if(PoolAddPosToList(hpS -> List[itm].Term, lpfy[k], (LPDWORD)&NewHead) == FALSE) return FALSE;
                        hpS -> List[itm].Term = NewHead;
                    }
                }
            }

            lk = lpPool[lk].Next;
        }
    }

    return TRUE;    
}

/* Returns the index in hpColl (collection of sets of items) if the set of
   items pointed to by hpSi is in hpColl, else -1. */
short int SetInCollection(
    LPSITEMS hpSi,          // Set of items
    WORD n                  // Number of sets in collection
    )
{
    register WORD i,j;
    LPLR1ILIST hpLr;

    for(i = 0; i < n; i++)                      // Loop on all sets of items of hpColl
    {
        if(hpSi -> Number != (hpCsiTable + i) -> Number) continue;

        for(j = 0; j < hpSi -> Number; j++)     // Loop on items of hpSi
        {
            hpLr = hpSi -> List + j;
            if(ItemIsIn(hpLr -> Prod, hpLr -> Pos, hpCsiTable + i) == -1) break;
        }
        if(j == hpSi -> Number) return i;
    }

    return -1;
}

/*** Write production to state file ***/
BOOL WriteProduction(WORD Prod)
{
    register WORD m;
    LPSYMBOL lpSymb;
    unsigned char memo;
    LPSTR Out;
    char tmpBuf[MAX_OUT_BUFFER + 1];

    lpSymb = hpSymbol + hpProduction[Prod].Left;
    memo = lpSymb -> hpSymbolStart[lpSymb -> SymbolSize];
    lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = '\0';
    CopyHugeString(tmpBuf, lpSymb -> hpSymbolStart);
    wsprintf((LPSTR)OutBuffer,"%u:\t%s -> ",Prod,(LPSTR)tmpBuf);
    lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = memo;
    Out = (LPSTR)OutBuffer; while(*Out) Out++;

    /* Loop on symbols at the right of the production, and write them to the state file */
    for(m = 0; m < hpProduction[Prod].Length; m++)
    {
        lpSymb = hpSymbol + hpProduction[Prod].Right[m];
        memo = lpSymb -> hpSymbolStart[lpSymb -> SymbolSize];
        lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = '\0';
    	CopyHugeString(Out, lpSymb -> hpSymbolStart);
        lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = memo;
        while(*Out) Out++;
        *Out++ = ' ';
    }

    *Out++ = '\r'; *Out++ = '\n';
    *Out = '\0';

    if(WriteToFile(hf,(LPSTR)OutBuffer)==FALSE) return FALSE;

    return TRUE;
}

BOOL Conflict(DWORD State, DWORD Symbol, LPSTR Act1, LPSTR Act2, WORD NewAct, WORD Act)
{
    char memo;
    char tmpBuf[MAX_OUT_BUFFER + 1];

    ConsultReport = TRUE;
    memo = hpSymbol[Symbol].hpSymbolStart[hpSymbol[Symbol].SymbolSize];
    hpSymbol[Symbol].hpSymbolStart[hpSymbol[Symbol].SymbolSize] = '\0';
    CopyHugeString(tmpBuf, hpSymbol[Symbol].hpSymbolStart);
    wsprintf((LPSTR)OutBuffer,"*** CONFLICT in state %lu, input token %s, %s %u <-> %s %u\r\n",
        State, (LPSTR)tmpBuf, Act1, NewAct, Act2, Act);
    hpSymbol[Symbol].hpSymbolStart[hpSymbol[Symbol].SymbolSize] = memo;
    if(WriteToFile(hf,(LPSTR)OutBuffer) == FALSE) return FALSE;
    if(PAction1 == REDUCE) if(WriteProduction(NewAct) == FALSE) return FALSE;
    if(PAction2 == REDUCE) if(WriteProduction(Act) == FALSE) return FALSE;
    if(WriteToFile(hf,(LPSTR)"\r\n") == FALSE) return FALSE;

    return TRUE;
}

/* Filling in the action table. Note that REDUCE 0 is not possible, since production
   0 is S' -> S. Reduce can then be represented by a negative number, and shift
   by a positive.
   Action conflicts are, resolved in the following way:
   1. Reduce / Reduce conflicts are resolved in favor of the production listed first in the
      source file.
   2. Shift / Reduce conflicts:
      2.1. If the production has a non zero precedence:
            2.1.1. If the precedence of the production is equal to that of the token
                2.1.1.1. If the associativity of the production differs from that of the token,
                            then there is an associativity mismatch error
                2.1.1.2. Otherwise, If the associativity if left, then reduce
                2.1.1.3. Else, shift
            2.1.3. Otherwise, if the precedence of the production is greater than that of the token, then reduce
            2.1.4. Else, shift
      2.2. Else, shift (handles the "dangling else" problem correctly)
*/
BOOL Fill(DWORD State, DWORD Symbol, WORD Action, WORD ActionType)
{
    short int Act,NewAct;
    char memo;
    DWORD lx;
    static LPSTR Reduce = "Reduce";
    static LPSTR Shift  = "Shift";
    LPSTR Act1,Act2;
    short int prec_prod,prec_tok;
    WORD Assoc,Assoc2;

    lx = (DWORD)nSymbol * (DWORD)State;
    NewAct = (short int)Action;
    PAction1 = ActionType;
    if(ActionType == REDUCE) Act1 = Reduce; else Act1 = Shift;
    Act = hpActionGoto[lx + Symbol];
    if(Act < 0) {Act = -Act; PAction2 = REDUCE; Act2 = Reduce;} else {PAction2 = SHIFT; Act2 = Shift;}
    if(hpActionGoto[lx + Symbol] != BLANK && (PAction1 != PAction2 || NewAct != Act))
    {
        /* A conflict */
        if(PAction1 != PAction2)    // Shift / Reduce conflict
        {
            if(hpProduction[NewAct].Precedence) // A precedence is defined for the production
            {
                prec_prod = hpProduction[NewAct].Precedence;
                prec_tok  = hpSymbol[Symbol].Precedence;
                Assoc = (prec_prod < 0 ? ASSOC_RIGHT : ASSOC_LEFT);
                Assoc2 = (prec_tok < 0 ? ASSOC_RIGHT : ASSOC_LEFT);
                if(prec_prod < 0) prec_prod = - prec_prod;
                if(prec_tok < 0)  prec_tok  = - prec_tok;

                if(prec_prod == prec_tok)
                {
                    if(Assoc != Assoc2)
                    {
                    	char tmpBuf[MAX_OUT_BUFFER + 1];
                    
                        /* Different associativities: associativity mismatch */
                        ConsultReport = TRUE;
                        memo = hpSymbol[Symbol].hpSymbolStart[hpSymbol[Symbol].SymbolSize];
                        hpSymbol[Symbol].hpSymbolStart[hpSymbol[Symbol].SymbolSize] = '\0';
                        CopyHugeString(tmpBuf, hpSymbol[Symbol].hpSymbolStart);
                        wsprintf((LPSTR)OutBuffer,"*** Associativity mismatch state %lu, input token %s, production %u\r\n",
                            State,(LPSTR)tmpBuf,NewAct);
                        hpSymbol[Symbol].hpSymbolStart[hpSymbol[Symbol].SymbolSize] = memo;
                        if(WriteToFile(hf,(LPSTR)OutBuffer) == FALSE) return FALSE;
                        return TRUE;    // Resolved in favor of Shift
                    }
                    else
                    {
                        if(Assoc == ASSOC_LEFT)
                            hpActionGoto[lx + Symbol] = - NewAct;   // Reduce
                        /* Else shift */
                    }
                }
                else    // Precedences are different
                {
                    if(prec_prod > prec_tok)
                        hpActionGoto[lx + Symbol] = - NewAct;   // Reduce
                    /* Else shift */
                }
            }
            else    // Else, shift
            {
                if(Conflict(State, Symbol, Act1, Act2, NewAct, Act) == FALSE) return FALSE;
            }
        }
        else                        // Reduce / Reduce conflict
        {
            if(Conflict(State, Symbol, Act1, Act2, NewAct, Act) == FALSE) return FALSE;
            if(NewAct < Act)
            {
                if(PAction1 == REDUCE) NewAct = -NewAct;
                hpActionGoto[lx + Symbol] = NewAct;
            }
        }
    }
    else
    {
        if(PAction1 == REDUCE) NewAct = -NewAct;
        hpActionGoto[lx + Symbol] = NewAct;
    }

    return TRUE;
}

/* Building the propagation table, and computing the lookaheads generated spontaneously */
BOOL PropaTable()
{
    register DWORD i,j,k,m,p;
    register DWORD lx;
    LPSITEMS lpS,lpSet;
    LPLR1ILIST hpLr,hpIt;
    WORD Prod,Pos,DestSet,Item;
    long int Terminal,GotoSet;

    for(i = 0; i < nCsi; i++)                   // Loop on each set of items I of hpCsiTable
    {
        lpS = hpCsiTable + i;
        lx = (DWORD)nSymbol * (DWORD)i;
        for(j = 0; j < lpS -> Number; j++)      // Loop on each item [B -> y.z] of the set
        {
            /* Eliminate item if the dot is at the end */
            hpLr = (lpS -> List) + j;
            if(hpLr -> Pos == hpProduction[hpLr -> Prod].Length) continue;

            /* Build the closure of the set limited to the item [B -> y.z] */
            if(Closure(i,j,(LPCSITEMS)&CIset) == FALSE) return FALSE;

            for(k = 0; k < CIset.Number; k++)   // Loop on items [A -> u.Xv,a] of the closure
            {
                Prod = CIset.List[k].Prod;
                Pos  = CIset.List[k].Pos;

                if(hpProduction[Prod].Length != 0)  // Not an empty production
                {
                    /* Take kernel item [A -> uX.v] of Goto(I,X), if exists */
                    GotoSet = hpActionGoto[lx + (DWORD)(hpProduction[Prod].Right[Pos])];
                    if(GotoSet == BLANK || GotoSet < 0) continue;   // No Goto Set for this item
                    DestSet = (WORD)GotoSet;
                    lpSet = hpCsiTable + DestSet;
                    Item = ItemIsIn(Prod, Pos + 1, lpSet);
                    hpIt = lpSet -> List + Item;
                }

                m = CIset.List[k].Term;
                while(m)    // Loop on the lookaheads of the couple (prod, pos)
                {
                    /* If Prod represents an empty production: [A -> empty .,a], with a <> P_EMTY, then
                       add the reduction immediately to the action table */
                    if(hpProduction[Prod].Length == 0)  // Empty production
                    {
                        if(lpPool[m].Pos != P_EMPTY)
                        {
                            if(Fill(i,lpPool[m].Pos,Prod,REDUCE) == FALSE) return FALSE;
                        }
                        else
                        {
                            /* Add the empty production to the special array: we must reduce by it on all
                               lookaheads computed for this state! */
                            if(lpS -> nEmptyProd < NEPROD)  // If already NEPROD productions registered, no use generating an error!
                                lpS -> EmptyProds[lpS -> nEmptyProd++] = Prod;
                        }
                    }
                    else
                    {
                        if(lpPool[m].Pos == P_EMPTY)
                        {
                            /* Propagation from [B -> y.z] to [A -> uX.v] */
                            if(AddPropa() == FALSE) return FALSE;
                            if(hpLr -> nPropa == 0) hpLr -> Propa = nPropa - 1;
                            hpLr -> nPropa++;
                            hpPropa[nPropa - 1].Set  = DestSet;
                            hpPropa[nPropa - 1].Item = Item;
                        }
                        else
                        {
                            /* Lookaheads generated spontaneously for [A -> uX.v] */
                            Terminal = lpPool[m].Pos;
                            p = hpIt -> Term;
                            while(p)
                            {
                                if(lpPool[p].Pos == Terminal) break;    // Terminal already present
                                p = lpPool[p].Next;
                            }

                            if(p == 0)
                            {
                                if(PoolAddPosToList(hpIt -> Term, Terminal, (LPDWORD)&NewHead) == FALSE) return FALSE;
                                hpIt -> Term = NewHead;
                            }
                        }
                    }

                    m = lpPool[m].Next;
                }
            }
        }
    }

    return TRUE;
}

/*** Write item to state file ***/
BOOL WriteItem(LPLR1ILIST hpIt)
{
    register WORD m;
    LPSYMBOL lpSymb;
    unsigned char memo;
    LPSTR Out;
    char tmpBuf[MAX_OUT_BUFFER + 1];

    lpSymb = hpSymbol + hpProduction[hpIt -> Prod].Left;
    memo = lpSymb -> hpSymbolStart[lpSymb -> SymbolSize];
    lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = '\0';
    CopyHugeString(tmpBuf, lpSymb -> hpSymbolStart);
    wsprintf((LPSTR)OutBuffer,"\t%u:\t%s -> ",hpIt -> Prod,(LPSTR)tmpBuf);
    lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = memo;
    Out = (LPSTR)OutBuffer; while(*Out) Out++;

    /* Loop on symbols at the right of the production, and write them to the state file */
    for(m = 0; m < hpProduction[hpIt -> Prod].Length; m++)
    {
        if(m == hpIt -> Pos)
        {
            /* Write the dot at current position in the production */
            *Out++ = '_';   // The dot
            *Out++ = ' ';
        }
        lpSymb = hpSymbol + hpProduction[hpIt -> Prod].Right[m];
        memo = lpSymb -> hpSymbolStart[lpSymb -> SymbolSize];
        lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = '\0';
    	CopyHugeString(Out, lpSymb -> hpSymbolStart);
        lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = memo;
        while(*Out) Out++;
        *Out++ = ' ';
    }

    /* Dot at the last position */
    if(m == hpIt -> Pos)
    {
        /* Write the dot at current position in the production */
        *Out++ = '_';   // The dot
        *Out++ = ' ';
    }

    *Out++ = '\r'; *Out++ = '\n';
    *Out = '\0';

    if(WriteToFile(hf,(LPSTR)OutBuffer)==FALSE) return FALSE;

    return TRUE;
}

/* Building the collection of sets of items of the augmented grammar */
BOOL SetsOfItems()
{
    register DWORD i,j,k,m,p;
    LPLR1ILIST hpLr,hpIt;
    BOOL Added;
    LPSITEMS lpS;
    long int Terminal;
    short int Dest;
    WORD NewCsi,iSymbol,Desc,Prod;
    char memo;
    LPSYMBOL lpSymb;
    char tmpBuf[MAX_OUT_BUFFER + 1];
    char tmpBuf2[MAX_OUT_BUFFER + 1];

    if(First(NULL,0,TRUE,NULL,NULL) == FALSE) return FALSE;     // Initialize First function
    if(ComputeDescendants() == FALSE) return FALSE;             // Compute descendants of each nonterminal

    /*** Build set I0, with unique kernel item [S' -> .S,$] ***/
    if(AddCsi() == FALSE) return FALSE;
    if(AddActionGoto() == FALSE) return FALSE;
    lpS = hpCsiTable;
    lpS -> Number = 1;
    hpLr = lpS -> List;
    hpLr -> Prod = 0;
    hpLr -> Pos = 0;
    hpLr -> Term = 0;

    /* The endmarker is generated spontaneously for unique item of I0. */
    if(PoolAddPosToList(hpLr -> Term, ENDMARKER, (LPDWORD)&NewHead) == FALSE) return FALSE;
    hpLr -> Term = NewHead;

    /* Write state I0 to state file */
    lpSymb = hpSymbol + hpProduction[0].Right[0];
    memo = lpSymb -> hpSymbolStart[lpSymb -> SymbolSize];
    lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = '\0';
    CopyHugeString(tmpBuf, hpSymbol[nTerm].hpSymbolStart);
    CopyHugeString(tmpBuf2, lpSymb -> hpSymbolStart);
    wsprintf((LPSTR)OutBuffer,"\r\nSTATE 0\r\n\t0:\t%s -> _ %s\r\n",(LPSTR)tmpBuf,(LPSTR)tmpBuf2);
    lpSymb -> hpSymbolStart[lpSymb -> SymbolSize] = memo;
    if(WriteToFile(hf,(LPSTR)OutBuffer)==FALSE) return FALSE;

    /* Build the kernel of the collection of sets of LR(0) items */
    for(i = 0; i < nCsi; i++)                   // Loop on each set of items I of hpCsiTable
    {
        lpS = hpCsiTable + i;
        for(j = 1; j < nSymbol; j++)            // Loop on symbols X of the grammar (excluding ENDMARKER)
        {
            /*** Build Goto(I,X) ***/

            if(j == nTerm) continue;            // Exclude also S'  (start symbol of the augmented grammar)

            Iset.Number = 0;                    // Initialize Goto(I,X)

            for(k = 0; k < lpS -> Number; k++)  // Loop on items: current is [D -> y.Cz]
            {
                /* Eliminate item if the dot is at the end */
                hpLr = (lpS -> List) + k;
                if(hpLr -> Pos == hpProduction[hpLr -> Prod].Length) continue;
    
                /* If C = X, add the kernel item [D -> yX.z] to Goto(I,X) */
                if((iSymbol = hpProduction[hpLr -> Prod].Right[hpLr -> Pos]) == j)
                {
                    if(ItemIsIn(hpLr -> Prod, hpLr -> Pos + 1, (LPSITEMS)&Iset) == -1)
                        if(AddItem(hpLr -> Prod, (hpLr -> Pos) + 1, (LPSITEMS)&Iset) == FALSE) return FALSE;
                }

                /* Add kernel items [A -> X.w] to Goto(I,X) such that A is a descendant from C
                   (A = C is also possible!) */
                if(iSymbol < nTerm) continue;                   // No descendant for a terminal
                for(m = 0; m < hpSymbol[iSymbol].nDesc; m++)    // Loop on descendants A of C
                {
                    Desc = hpSymbol[iSymbol].Descendants[m];
                    for(p = 0; p < hpSymbol[Desc].nLeftProd; p++)   // Loop on productions having A as left side
                    {
                        Prod = hpSymbol[Desc].LeftProd[p];
                        if(hpProduction[Prod].Length > 0 &&
                           hpProduction[Prod].Right[0] == j)        // X is the first symbol on the right side
                        {
                            if(ItemIsIn(Prod, 1, (LPSITEMS)&Iset) == -1)
                                if(AddItem(Prod, 1, (LPSITEMS)&Iset) == FALSE) return FALSE;
                        }
                    }
                }
            }

            if(Iset.Number == 0) continue;                      // Do not add an empty set!

            if((Dest = SetInCollection((LPSITEMS)&Iset, nCsi)) == -1)
            {
                /*** Add the set Goto(I,X) to hpCsiTable ***/
                if(AddCsi() == FALSE) return FALSE;
                NewCsi = nCsi - 1;

                /* Write state number to state file */
                wsprintf((LPSTR)OutBuffer,"\r\nSTATE %u\r\n",NewCsi);
                if(WriteToFile(hf,(LPSTR)OutBuffer)==FALSE) return FALSE;

                lpS = hpCsiTable + i;   // Important, because memory reallocation can cause pointer to change!
                (hpCsiTable + NewCsi) -> Number = Iset.Number;
                for(k = 0; k < Iset.Number; k++)
                {
                    hpLr = Iset.List + k;
                    hpIt = (hpCsiTable + NewCsi) -> List + k;
                    hpIt -> Prod   = hpLr -> Prod;
                    hpIt -> Pos    = hpLr -> Pos;
                    hpIt -> Term  = 0;

                    /*** Write item to state file ***/
                    if(WriteItem(hpIt) == FALSE) return FALSE;
                }
                Dest = NewCsi;
                if(AddActionGoto() == FALSE) return FALSE;
            }

            if(Fill(i,j,Dest,SHIFT) == FALSE) return FALSE;
        }
    }

    if(WriteToFile(hf,(LPSTR)"\r\n")==FALSE) return FALSE;

    /* Building the propagation table, and computing the lookaheads generated spontaneously */
    if(PropaTable() == FALSE) return FALSE;

    /* Make repeated passes over the kernel items, to propagate lookaheads */
    do
    {
        Added = FALSE;
        for(i = 0; i < nCsi; i++)               // Loop on the sets of items
        {
            lpS = hpCsiTable + i;
            for(j = 0; j < lpS -> Number; j++)  // Loop on the items of the current set
            {
                hpLr = lpS -> List + j;

                /* Propagate lookaheads of item to related items */
                for(k = 0; k < hpLr -> nPropa; k++)
                {
                    hpIt = hpCsiTable[hpPropa[hpLr -> Propa + k].Set].List + hpPropa[hpLr -> Propa + k].Item;

                    m = hpLr -> Term;
                    while(m)        // Loop on the terminals to propagate
                    {
                        Terminal = lpPool[m].Pos;
                        p = hpIt -> Term;
                        while(p)
                        {
                            if(lpPool[p].Pos == Terminal) break;    // Terminal already present
                            p = lpPool[p].Next;
                        }

                        if(p == 0)
                        {
                            if(PoolAddPosToList(hpIt -> Term, Terminal, (LPDWORD)&NewHead) == FALSE) return FALSE;
                            hpIt -> Term = NewHead;
                            Added = TRUE;
                        }

                        m = lpPool[m].Next;
                    }
                }
            }
        }
    } while(Added == TRUE);

    return TRUE;
}

/* Building the Action - Goto table for an augmented grammar */
BOOL BuildParsingTable(LPSTR TargetDir)
{
    register DWORD i,j,k,m,p,q;
    LPLR1ILIST hpLr;
    LPSITEMS hpS;
    WORD Prod,iSymb,Desc,EmptyProd;
    OFSTRUCT ofs;
    unsigned char memo;
	char tmpBuf[MAX_OUT_BUFFER + 1];

    ConsultReport = FALSE;

    lstrcpy((LPSTR)StateFile,TargetDir);
    lstrcat((LPSTR)StateFile,"\\Parser.txt");
    hf = OpenFile((LPSTR)StateFile, &ofs, OF_WRITE|OF_CREATE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_OPEN_FILE;
        return FALSE;
    }

    LoadString(hDLLInst,IDS_DO_NOT_EDIT,(LPSTR)OutBuffer,MAX_OUT_BUFFER);
    if(WriteToFile(hf,(LPSTR)OutBuffer)==FALSE) return FALSE;
    if(WriteToFile(hf,(LPSTR)"/* Parser Construction Report File */\r\n")==FALSE) return FALSE;

    /* Dynamic arrays used for computations of closure */
    lpy =  (long int HOST_FAR *)GlobalAllocPtr(GHND,(DWORD)sizeof(long int) * (DWORD)LENPROD);
    lpfy = (long int HOST_FAR *)GlobalAllocPtr(GHND,(DWORD)sizeof(long int) * (DWORD)NTERM);
    if(lpy == NULL || lpfy == NULL) {LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR; return FALSE;}

    if(SetsOfItems() == FALSE) {_lclose(hf); return FALSE;}     // Build collection of sets of LALR(1) items, and the goto table

    for(i = 0; i < nCsi; i++)                                   // Loop on all states
    {
        for(j = 0; j < (hpCsiTable + i) -> Number; j++) // Loop on all items of the state
        {
            hpS = hpCsiTable + i;
            hpLr = hpS -> List + j;
            Prod = hpLr -> Prod;
            k = hpLr -> Term;
            while(k)                    // Loop on all lookahead terminals of the item
            {
                /* Note: the shift actions were computed while building the Goto transitions for terminals. */

                if(Prod == 0 && hpLr -> Pos == 1 && lpPool[k].Pos == ENDMARKER)
                {                                               // Braces to avoid attaching the else to the wrong if!
                    /* ACCEPT */
                    if(Fill(i,ENDMARKER,ACCEPT,SHIFT) == FALSE) {_lclose(hf); return FALSE;}
                }
                else
                {
                    if(hpLr -> Pos == hpProduction[Prod].Length && hpProduction[Prod].Left != nTerm)
                    {                                           // Braces to avoid attaching the else to the wrong if!
                        /* REDUCE */
                        if(Fill(i,lpPool[k].Pos,Prod,REDUCE) == FALSE) {_lclose(hf); return FALSE;}
                    }
                    else
                    {
                        if(hpLr -> Pos < hpProduction[Prod].Length && (iSymb = hpProduction[Prod].Right[hpLr -> Pos]) >= nTerm)
                        {
                            /* The dot is not at the end of the production, and the symbol at the right of the dot
                               is a nonterminal ! Special case: reduction by an empty production A -> empty
                               Reduce by the empty production, on all lookaheads listed
                                 in the Term list of the item! */

                            for(m = 0; m < hpSymbol[iSymb].nDesc; m++)          // Loop on the decsendants of the nonterminal
                            {
                                Desc = hpSymbol[iSymb].Descendants[m];          // Current descendant: A
                                for(p = 0; p < hpSymbol[Desc].nLeftProd; p++)   // Loop on productions for which A is the left side
                                {
                                    /* Reject any production other than empty productions */
                                    EmptyProd = hpSymbol[Desc].LeftProd[p];
                                    if(hpProduction[EmptyProd].Length != 0) continue;

                                    /* The production must be in the special list */
                                    for(q = 0; q < hpS -> nEmptyProd; q++)
                                        if(EmptyProd == hpS -> EmptyProds[q]) break;
                                    if(q < hpS -> nEmptyProd)
                                    {
                                        /* Add the REDUCE action for this production */
                                        if(Fill(i,lpPool[k].Pos,EmptyProd,REDUCE) == FALSE) {_lclose(hf); return FALSE;}
                                    }
                                }
                            }
                        }
                    }
                }

                k = lpPool[k].Next;
            }
        }
    }

    GlobalFreePtr(lpy); GlobalFreePtr(lpfy);

    /* Check that each nonterminal is on the left side of at least one production! */
    for(i = nTerm + 1; i < nSymbol; i++)
    {
        if(hpSymbol[i].nLeftProd == 0)
        {
            memo = hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize];
            hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize] = '\0';
            CopyHugeString(tmpBuf, hpSymbol[i].hpSymbolStart);
            wsprintf((LPSTR)OutBuffer,"*** ERROR: Nonterminal '%s' is on the left side of NO production!\r\n",
                (LPSTR)tmpBuf);
            hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize] = memo;
            if(WriteToFile(hf,(LPSTR)OutBuffer)==FALSE) return FALSE;
            ConsultReport = TRUE;
        }
    }

    /* Check whether some tokens are unused */
    if(WriteToFile(hf,(LPSTR)"\r\n") == FALSE) return FALSE;
    for(i = 2; i < nTerm; i++)
    {
        if(hpSymbol[i].TermUsed == FALSE)
        {
            unsigned char c;

            ConsultReport = TRUE;
            c = hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize];
            hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize] = '\0';
            CopyHugeString(tmpBuf, hpSymbol[i].hpSymbolStart);
            wsprintf((LPSTR)OutBuffer,"*** WARNING: Token '%s' does not figure in any production\r\n",
                (LPSTR)tmpBuf);
            hpSymbol[i].hpSymbolStart[hpSymbol[i].SymbolSize] = c;
            if(WriteToFile(hf,(LPSTR)OutBuffer) == FALSE) return FALSE;
        }
    }

    _lclose(hf);

    return TRUE;
}

