                            /**********************************************
                            * Phoenix: WINDOWS Parser Generator           *
                            **********************************************/
 /********                      HUGE ARRAYS
 ********* !!!!!!!!! CAUTION: THIS PACKAGE USES SOME HUGE POINTERS, AND HUGE ARRAYS:
 ********* !!!!!!!!! REMEMBER THAT A HUGE ARRAY ELEMENT, UNDER WINDOWS 3.1X, MAY
 ********* !!!!!!!!! NEVER CROSS A SEGMENT (64K) BOUNDARY. THIS IMPLIES THAT ALL
 ********* !!!!!!!!! ELEMENTS OF A HUGE ARRAY MUST OCCUPY A NUMBER OF BYTES EQUAL TO
 ********* !!!!!!!!! A POWER OF 2, AND INFERIOR TO 64K !
 ********* !!!!!!!!! WHEN USING HUGE ARRAYS OF STRUCTURES, MAKE SURE TO SET THE
 ********* !!!!!!!!! "Struct Member Byte Alignment" COMPILATION OPTION TO 1 BYTE. THIS WAY,
 ********* !!!!!!!!! YOU CAN EASILY FIT THE SIZE OF THE STRUCTURE TO A POWER OF 2
 ********* !!!!!!!!! WITH A PADDING FIELD.
 *********/

#include "ph_sys.h"

/* Macros used in symbol installation into the symbol table hpSymbol */
#define SYMBOL_TOKEN        0
#define SYMBOL_NONTERMINAL  1
#define SYMBOL_UNDETERMINED 2

/* Special tokens */
#define ENDMARKER   0                           // The endmarker terminal
#define ERROR_TOKEN 1                           // The error virtual token, used for error productions

/* Parser actions */
#define SHIFT       1                           // The shift action of the Parser
#define REDUCE      2                           // The reduce action of the Parser
#define ACCEPT      32767                       // The "accept" action
#define BLANK       (-32768)                    // The initial value of each entry of hpActionGoto

#define NTERM       128                         // Maximum number of terminals in a production
#define LENPROD     32                          // Maximum number of symbols on the right of a production
#define P_EMPTY     (-1)                        // The "empty" terminal. Does NOT exist in hpSymbol as a terminal,
                                                // rather is the virtual -1 index to hpSymbol.
#define COUPLES     62                          // Maximum number of (production, position) couples in a
                                                // set of items (kernels of a set of items).
#define MAX_ITEMS   510                         // Maximum number of items in the set used for closure computation
#define NLEFTPROD   100                         // Maximum number of productions with the same left side nonterminal
#define NDESC       130                         // Maximum number of descendants of a nonterminal
#define NEPROD      10                          // Maximum number of conflicting empty productions in the same state

/* Associativity for tokens or productions */
#define ASSOC_NONE  0
#define ASSOC_LEFT  1
#define ASSOC_RIGHT 2


                            /*** Grammar symbols ***/

#define SYMBOL_INCREMENT    100                 // Increment for symbol allocation

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _SYMBOLS                         // Type for X and First(X), X being a grammar symbol
{
    BOOL        TermUsed;                       // Indicates whether the symbol, if a terminal, is used in at least one
                                                // production, or not.
    short int   Precedence;                     // Precedence order. Negative if right associative, positive if left associative,
                                                // zero if no precedence order.
    WORD        nTerm;                          // Number of entries in the array below
    short int   First[NTERM];                   // Terminals (indexes to lpSymbol[] itself). May be negative,
                                                // because may be EMPTY (the virtual -1 index to hpSymbol)!
    WORD        nLeftProd;                      // Number of entries in the array below
    WORD        LeftProd[NLEFTPROD];            // List of indexes to productions having the current symbol as left side
                                                // (for nonterminals only!)
    WORD        nDesc;                          // Number of entries in the array below
    WORD        Descendants[NDESC];             // List of indexes to nonterminals A such as C => As with a rightmost derivation,
                                                // C being the current symbol and a nonterminal!
    LPHUGE      hpSymbolStart;                  // Pointer to begining of symbol identifier
    DWORD       SymbolSize;                     // Number of characters of symbol identifier
#ifdef WIN31
    /* The following structure field is not used in the program, but is essential to complete
       the size of the structure to a power of 2 (see note above on huge arrays) */
    char  padder[348];
#endif
} SYMBOLS;
typedef SYMBOLS HOST_HUGE *LPSYMBOL;            // Pointer type to a symbol structure
LPSYMBOL hpSymbol;                              // Symbols used in the grammar: terminals are listed first,
                                                // then nonterminals. The first nonterminal listed is the
                                                // added nonterminal of the augmented grammar. For example,
                                                // if S is the start symbol, then the grammar is augmented
                                                // with production S' -> S, and S' is the first nonterminal
                                                // listed. The first terminal listed is ENDMARKER, and the
                                                // second is the ERROR_TOKEN.
WORD nSymbol;                                   // Current number of symbols in the array lpSymbol[].
WORD nMaxSymbol;                                // Maximum number of symbols in the array lpSymbol[].
WORD nTerm;                                     // Current number of terminals



                            /*** Productions. S being the start symbol, s' -> S is the
                                 first production. ***/

#define PRODUCTION_INCREMENT    100             // Increment for production allocation

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _PROD
{
    short int   Precedence;                     // Precedence order. Negative if right associative, positive if left associative,
                                                // zero if no precedence order.
    WORD        Left;                           // Index in lpSymbol[] of the nonterminal to the left of the
                                                // production.
    WORD        Length;                         // Number of symbols on the right of the production; a length of zero
                                                // indicates an empty production.
    WORD        Right[LENPROD];                 // Right side of the production: indexes to lpSymbol[]
    LPHUGE      hpActionStart;                  // Begining of code of associated action. NULL pointer if no action.
    DWORD       ActionSize;                     // Size of action.
    DWORD       ActionLineNumber;               // Source line where action starts (for #line directive generation)
#ifdef WIN31
    /* The following structure field is not used in the program, but is essential to complete
       the size of the structure to a power of 2 (see note above on huge arrays) */
    char  padder[46];
#endif
} PRODUCTION;

typedef PRODUCTION HOST_HUGE *LPPRODUCTION;     // Pointer type to a production
LPPRODUCTION hpProduction;                      // Dynamic array of productions
WORD nProduction;                               // Current number of productions in hpProduction[].
WORD nMaxProduction;                            // Maximum number of productions in hpProduction[].


                            /*** Collection of sets of items, LR(1) ***/

#define CSI_INCREMENT           100             // Increment for sets of items

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _LR1_ITEM_LIST                   // LR(1) items of the form (Prod, Pos, Term(i))
{
    WORD        Prod;                           // Production index in hpProduction[]
    WORD        Pos;                            // Position of the dot in the production (starts at 0)
    WORD        Propa;                          // Index to the propagation table
    WORD        nPropa;                         // Number of target items in the propagation list
    DWORD       Term;                           // List of terminals for this couple (prod, pos). May include
                                                // P_EMPTY (the virtual -1 index to hpSymbol)! Term is an index in
                                                // lpPool.
                                                // These terminals are the item lookaheads.
#ifdef WIN31
    /* The following structure field is not used in the program, but is essential to complete
       the size of the structure to a power of 2 (see note above on huge arrays) */
    char  padder[4];
#endif
} LR1_ITEM_LIST;
typedef LR1_ITEM_LIST HOST_HUGE *LPLR1ILIST;    // Pointer type to a LR1_ITEM_LIST structure

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _SET_OF_ITEMS
{
    WORD            nEmptyProd;                 // Number of elements in the following array
    WORD            EmptyProds[NEPROD];         // Empty productions by which to reduce on all computed lookaheads:
                                                // this comes from item [A -> empty,P_EMPTY] computed in closure()!
                                                // Note: if there is more than 1 production in this array, then
                                                // reduce - reduce conflicts will be reported!
    WORD            Number;                     // Number of (prod, pos) couples in this set
    LR1_ITEM_LIST   List[COUPLES];              // Set of items
#ifdef WIN31
    /* The following structure field is not used in the program, but is essential to complete
       the size of the structure to a power of 2 (see note above on huge arrays) */
    char  padder[8];
#endif
} SET_OF_ITEMS;

typedef SET_OF_ITEMS HOST_HUGE *LPSITEMS;       // Pointer type to a set of items
LPSITEMS hpCsiTable;                            // Collection of sets of items (dynamic array)
WORD nCsi;                                      // Current number of sets of items
WORD nMaxCsi;                                   // Maximum number of sets of items

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _CLOSED_SET_OF_ITEMS             // Structure to compute closed sets of items
{
    WORD            nEmptyProd;                 // Number of elements in the following array
    WORD            EmptyProds[NEPROD];         // Empty productions by which to reduce on all computed lookaheads:
                                                // this comes from item [A -> empty,P_EMPTY] computed in closure()!
                                                // Note: if there is more than 1 production in this array, then
                                                // reduce - reduce conflicts will be reported!
    WORD            Number;                     // Number of (prod, pos) couples in this set
    LR1_ITEM_LIST   List[MAX_ITEMS];            // Set of items
#ifdef WIN31
    /* The following structure field is not used in the program, but is essential to complete
       the size of the structure to a power of 2 (see note above on huge arrays) */
    char  padder[8];
#endif
} CLOSED_SET_OF_ITEMS;
typedef CLOSED_SET_OF_ITEMS HOST_HUGE *LPCSITEMS;   // Pointer type to a set of items

#define PROPA_INCREMENT         1000            // Increment for propagation table

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _PROPAGATION
{
    WORD    Set;                                // Index to a set of items in hpCsiTable
    WORD    Item;                               // Index to an item in List[] of the current set
} PROPAGATION;
typedef PROPAGATION HOST_HUGE *LPPROPA;         // Pointer type to the propagation table
LPPROPA hpPropa;                                // The dynamic propagation table. Each item has an index
                                                // to this table indicating the begining of the list of items
                                                // to which the lookaheads propagate.
WORD nPropa;                                    // Current number of entries in the propagation table
WORD nMaxPropa;                                 // Maximum number of entries in the propagation table


                            /*** Action - Goto table ***/

#define ACTION_GOTO_INCREMENT   100             // Increment for action-goto allocation
typedef short int HOST_HUGE *LPACTIONGOTO;      // The action-goto table type
LPACTIONGOTO hpActionGoto;                      // The action-goto table.
                                                // Action and Goto functions for a state:
                                                // action is defined for the terminals,
                                                // goto for the nonterminals.
                                                // Action:
                                                // > 0 means shift
                                                // < 0 means reduce
                                                // = 0 means error
                                                // ACCEPT means accept.
                                                // Goto: the index of the state to go to.
WORD nActionGoto;                               // Current number of lines in the action-goto table. The number of entries
                                                // is: nActionGoto * nSymbol
WORD nMaxActionGoto;                            // Maximum number of lines in the action-goto table. The maximum number
                                                // of entries is nMaxActionGoto * nSymbol
BOOL ConsultReport;                             // Flag set to TRUE after a parser generation,
                                                // if at least 1 conflict,error or warning was generated
/* Packed Action and Goto arrays.
   -----------------------------

    There are 3 arrays: ActionState, GotoSymbol and AGPool.
    The 2 first are arrays of indexes to contiguous lists stored in AGPool.
    ActionState has an entry for each state.
    GotoSymbol has an entry for each nonterminal (except S', the start symbol of the augmented grammar).
    AGPool is an array of double words of 2 bytes. The first word contains the symbol (action) or the current state (goto).
    The second word contains the action to perform or the state to go to.
    The end of each list is indicated by the value END_OF_LIST. For the action part, an END_OF_LIST, when reached,
    indicates a syntax error (BLANK). For the goto part, it indicates the default goto value (the most frequent).
*/
#define END_OF_LIST             (-1)
#define AGPOOL_INCREMENT        10000
short int HOST_FAR *ActionState;                // The ActionState dynamic array.
LPWORD GotoSymbol;                              // The GotoSymbol dynamic array
DWORD HOST_HUGE *AGPool;                        // The AGPool dynamic array
WORD nAGPool;                                   // The current number of entries in AGPool
WORD nMaxAGPool;                                // The maximum number of entries in AGPool

                            /*** Internal Procedures and Functions ***/

BOOL    PInit               ();                 // Parser generator Initializations
BOOL    PEnd                ();                 // Free dynamic arrays
BOOL    BuildParsingTable   ();                 // Building the Action - Goto table for an augmented grammar S' -> S
BOOL    ParserPrepare       ();                 // Analyze the source parser file, and store data in intermediate structures
BOOL    AddSymbol           ();                 // Create dynamic space for new grammar symbol
BOOL    AddProduction       ();                 // Create dynamic space for new production
void    FreePackedArrays    ();                 // Free memory allocated for packed arrays construction
BOOL    PackActionGoto      ();                 // Pack Action and Goto tables
