                            /**********************************************
                            * Phoenix: WINDOWS Lexical Analyzer Generator *
                            **********************************************/

 /********                      HUGE ARRAYS
 ********* !!!!!!!!! CAUTION: THIS PACKAGE USES SOME HUGE POINTERS, AND HUGE ARRAYS:
 ********* !!!!!!!!! REMEMBER THAT A HUGE ARRAY ELEMENT, UNDER WINDOWS 3.1X, MAY
 ********* !!!!!!!!! NEVER CROSS A SEGMENT (64K) BOUNDARY. THIS IMPLIES THAT ALL
 ********* !!!!!!!!! ELEMENTS OF A HUGE ARRAY MUST OCCUPY A NUMBER OF BYTES EQUAL TO
 ********* !!!!!!!!! A POWER OF 2, AND INFERIOR TO 64K !
 ********* !!!!!!!!! WHEN USING HUGE ARRAYS OF STRUCTURES, MAKE SURE TO SET THE
 ********* !!!!!!!!! "Struct Member Byte Alignment" COMPILATION OPTION TO 1 BYTE. THIS WAY,
 ********* !!!!!!!!! YOU CAN EASILY FIT THE SIZE OF THE STRUCTURE TO A POWER OF 2
 ********* !!!!!!!!! WITH A PADDING FIELD.
 *********/

#include "ph_sys.h"

                                    /**** Datatypes and variables ***/

#define MAX_CHARS       256                     // Maximum number of characters in the alphabet
BOOL CaseSensitive;                             // Variable determining case sensitivity or regular expressions. Default: TRUE

/* Special character definitions */
#define EOI             0                       // End Of Input character
#define EMPTY           MAX_CHARS + 0           // Empty character (@)
#define CAT             MAX_CHARS + 1           // Concatenation character. Does not exist in source expressions.
                                                // Rather, is generated by regular expression anylyzer.
#define STAR            MAX_CHARS + 2           // * operator: zero or n times an expression
#define OR              MAX_CHARS + 3           // | operator: expression1 or expression2
#define LEFT            MAX_CHARS + 4           // left parenthesis
#define RIGHT           MAX_CHARS + 5           // right parenthesis
#define QMARK           MAX_CHARS + 6           // ? operator: zero or one time an expression
#define PLUS            MAX_CHARS + 7           // + operator: one or n times an expression

/* A pool is used to compute the functions firstpos, lastpos and followpos, to optimize memory use (avoid
   waste), and get rid of limits, the pool increasing dynamically as needed.
   In the pool are stored linked sets of positions.
   The pool is also used by the parser generator, to store item lookaheads, saving considerable space. */
#define POOL_INCREMENT      10000               // Increment for memory allocation of pool

/* Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _POOL                            // Type of a pool entry
{
    long int    Pos;                            // A position in the syntax tree of the regular expression
    DWORD       Next;                           // Index to the next position in the set. 0 means "end of set".
} POOL;
typedef POOL HOST_HUGE *LPPOOL;                 // Type of pool array
LPPOOL lpPool;                                  // The pool array. Entry 0 is unused.
DWORD nPool;                                    // The current top of the pool (the last used entry)
DWORD nMaxPool;                                 // Maximum number of entries in the pool
DWORD Holes;                                    // Index to the linked set of holes to reuse in the pool. Value
                                                // zero means "no hole".

/* The stack used for depth-first traversals of the syntax tree of the regular expression.
   Index 0 of stack array is always unused. */
#define STACK_INCREMENT     500                 // Increment for memory allocation of array lpStack
typedef struct _STACK                           // Type of a stack entry
{
    BOOL    Nullable;                           // Nullable function. Values: TRUE or FALSE
    DWORD   Firstpos;                           // Firstpos function: index to a set of leaf positions in the pool
    DWORD   Lastpos;                            // Lastpos  function: index to a set of leaf positions in the pool
} STACK;
typedef STACK HOST_FAR *LPSTACK;                // Type of stack array
LPSTACK lpStack;                                // The stack array. Entry 0 is unused.
DWORD StackIndex;                               // Index of current stack top (the last used entry)
DWORD nMaxStack;                                // Maximum number of entries in the stack

#define POSITION_INCREMENT      1000            // Increment for memory allocation of array lpPosition
typedef struct _POSITION                        // Type of an entry of position array
{
    WORD    Label;                              // The character value at this position
    DWORD   Followpos;                          // For followpos function. Index to a list of positions in the pool
} POSITION;
typedef POSITION HOST_FAR *LPPOSITION;          // Type of position array
LPPOSITION lpPos;                               // The position array. Entry 0 is unused.
DWORD nPos;                                     // Current number of used entries in lpPos
DWORD nMaxPos;                                  // Maximum number of entries in lpPos

/* Character classes, like for example [A-Za-z]. They are treated as characters while DFA construction, and finally
   expanded after DFA is built */
#define MAX_CLASSES                 50          // Maximum number of distinct classes among all expressions together

/* Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef BYTE ONE_CLASS[MAX_CHARS];              // Definition of a class: if character c is in the class, then entry c is != 0
typedef ONE_CLASS HOST_HUGE *LPCLASS;
LPCLASS hpClass;                                // The class dynamic array
WORD nClass;                                    // Current number of classes in hpClass

WORD Alphabet[MAX_CHARS + MAX_CLASSES];         // Alphabet. There are Alphabet[0] characters in the alphabet

/* State table */
#define LEXSTATE_INCREMENT          100         // Increment for memory allocation of array lpStates
typedef struct _LEXSTATE                        // Definition of a state
{
    DWORD   Subset;                             // Subset of positions. An index to a linked list in the pool
    WORD    Token;                              // Recognized token
} LEXSTATE;
typedef LEXSTATE HOST_FAR *LPLEXSTATE;          // Far pointer to dynamic array of states
LPLEXSTATE lpStates;                            // Dynamic array of states. lpStates[0] unused. For each regular expression,
                                                // the array is cleared and used again to construct a DFA.
WORD nStates;                                   // Last state index in lpStates. The real current number of states is:
                                                // nStates + 1, since lpStates[0] is unused!
WORD nMaxStates;                                // Maximum number of states in lpStates[]

/* Transition tables */
#define LEXTRANSITION_INCREMENT     100         // Increment for memory allocation of transition tables

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _LEXTRANSITION
{
    WORD    NextState[MAX_CHARS];               // Transition array for the state: each entry is an index to the table itself
    WORD    NextStateClass[MAX_CLASSES];        // Transition array for the state, on classes: each entry is an index to the
                                                // table itself, and the index of this array is the index to hpClass.
    WORD    Group;                              // Group number, for partitionment algorithm, while minimizing DFA
    WORD    Token;                              // Recognized token
#ifdef WIN31
    /* The following structure field is not used in the program, but is essential to complete
       the size of the structure to a power of 2 (see note above on huge arrays) */
    char  padder[408];
#endif
} LEXTRANSITION;
typedef LEXTRANSITION HOST_HUGE *LPLEXTRANSITION;   // Huge pointer to dynamic array of transitions
LPLEXTRANSITION hpGNextState,                   // Transition table used while construction. The states built after
                                                // each DFA construction on a regular expression are appended to this
                                                // table.
                hpFNextState;                   // Second transition table used while building global DFA
WORD TotalStates;                               // Current number of entries in hpGNextState, without counting current
                                                // construction. Thus, the real current number of entries is:
                                                // TotalStates + nNextState + 1, since hpGNextState[0][] is unused!
WORD nMaxGNextState;                            // Maximum number of states in hpGNextState
WORD nMaxFNextState;                            // Maximum number of states in hpFNextState
DWORD u;                                        // Set of positions used during construction. Index to linked list in the pool
WORD DfaStart;                                  // First state of the DFA
WORD nToken;                                    // Token counter, NOT set to zero after each DFA construction

    /**** Types and variables to store data after reading input buffer ****/
    
typedef unsigned char HOST_HUGE *LPHUGE;        // Huge pointer type
typedef struct _LEXCODEPIECE                    // Type for piece of code in the input buffer
{
    LPHUGE      hpCodeStart;                    // Pointer to beginning of code. First character must be {
    long int    CodeSize;                       // Number of characters of code fragment, including enclosing { }.
                                                // When CodeSize is negative, it represents the action "return token_value",
                                                // where the token value to return is stored in hpCodeStart.
    DWORD       LineNumber;                     // Source line where code starts (for #line directive generation)
} LEXCODEPIECE;

/* PART 1 */
#define MAX_REG_EXP                 400         // Maximum number of characters in a regular expression (trailing zero included)
#define CLASS_CODES                 16000       // Base for class codes
#define LEXP1DECL_INCREMENT         50          // Increment for memory allocation of array of declarations of Part 1
typedef WORD HOST_HUGE *HPWORD;                 // WORD huge pointer type
typedef struct _LEXP1DECL                       // Type for identifier declaration in part one, in the input buffer
{
    LPHUGE  lp1IdStart;                         // Pointer to identifier in input buffer
    WORD    IdSize;                             // Size of identifier
    WORD    RegExp[MAX_REG_EXP];                // Equivalent regular expression (recursively expanded). Classes are replaced
                                                // by special class codes (indexes to hpClass, + CLASS_CODES)
    WORD    RegExpSize;                         // Size of regular expression
} LEXP1DECL;
typedef LEXP1DECL HOST_FAR *LPLEXP1DECL;        // Far pointer to dynamic array of declarations in Part 1
LEXCODEPIECE P1Code;                            // Code in Part 1
LPLEXP1DECL lpP1Decl;                           // Pointer to list of declarations of Part 1
WORD nP1Decl;                                   // Number of declarations in Part 1
WORD nMaxP1Decl;                                // Maximum number of declarations in Part 1

/* PART 2 */
#define LEXP2REGEXP_INCREMENT       50          // Increment for memory allocation of array of reg. exp. of Part 2

/* WIN31: Make sure that the size of the following
   structure is a power of 2 (see note above
   on huge arrays). */
typedef struct _LEXP2REGEXP                     // Type for regular expression in the input buffer
{
    WORD            RegExp[MAX_REG_EXP];        // Regular expression (recursively expanded), enclosed in ( ), suffixed by #,
                                                // and with trailing zero. Classes are replaced by special class codes (indexes
                                                // to hpClass, + CLASS_CODES)
    WORD            RegExpSize;                 // Size of regular expression
    DWORD           RegExpLineNumber;           // Source line where regular expression starts (for error messages)
    DWORD           RegExpColumnNumber;         // Source column where regular expression starts (for error messages)
    LEXCODEPIECE    Action;                     // Action associated with regular expression
#ifdef WIN31
    /* The following structure field is not used in the program, but is essential to complete
       the size of the structure to a power of 2 (see note above on huge arrays) */
    char  padder[202];
#endif
} LEXP2REGEXP;
typedef LEXP2REGEXP HOST_HUGE *HPLEXP2REGEXP;   // Huge pointer to dynamic array of regular expressions in Part 2
HPLEXP2REGEXP hpP2RegExp;                       // Pointer to list of regular expressions of Part 2
WORD nP2RegExp;                                 // Number of regular expressions in Part 2
WORD nMaxP2RegExp;                              // Maximum number of regular expressions in Part 2

/* PART 3 */
LEXCODEPIECE P3Code;                            // Pointer to code in Part 3

/* Input buffer */
LPHUGE hpInputBuffer;                           // Pointer to input buffer
long InputSize;                                 // Size of the input buffer (trailing zero excluded)
DWORD yyline;                                   // The current line number
DWORD yycol;                                    // The current column number

/* Output buffer */
#define MAX_OUT_BUFFER              256         // Maximum length of output buffer (trailing zero excluded)
#define MAX_FILE_NAME               128         // Maximum length of file name
unsigned char OutBuffer[MAX_OUT_BUFFER + 1];
unsigned char OutParam [MAX_OUT_BUFFER + 1];

/* DFA Packing */
#define CHECK_INCREMENT             10000
#define VIRGIN                      32767       // Value used to mark a position as 'virgin' in the PCheck array
LPWORD PBase,PDefault,PNext,PCheck;             // Arrays for packed DFA
WORD nCheck;                                    // Length of PNext and PCheck arrays
WORD nMaxCheck;                                 // Maximum length of PNext and PCheck arrays
WORD Percent;                                   // Percentage of matching gotos in two states, over which decision is taken to
                                                // merge the states

                                    /* Internal functions: miscellaneous */

void    CopyHugeString  (LPSTR lpDest, LPHUGE hpSource); /* Copy a null terminated string */
void    CopyHugeStringN (LPSTR lpDest, LPHUGE hpSource, unsigned int uiLength); /* Copy a non-null terminated string. No null terminator is appended. */
BOOL    LexInit         ();                 // Initialize lex variables
BOOL    LexEnd          ();                 // Memory management to do before exiting program
BOOL    LoadFile        (LPSTR FileName);   // Read the input file; store it into the input buffer
BOOL    LexPrepare      ();                 // Prepare the data structures with data from input buffer
void    LexError        (HWND hwnd);        // Manage Phoenix errors
BOOL    DfaInit         ();                 // Initialize variables for global DFA construction
void    DFAFree         ();                 // Free dynamic arrays used for DFA construction
BOOL    GetSourceToken  ();                 // Read the next source token
BOOL    Regular         ();                 // Analyze syntax of regular expression, build syntax tree
BOOL    Tables          ();                 // Build DFA from regular expression's Followpos[][] table
BOOL    AddClass        ();                 // Replace character classes by their equivalent characters
BOOL    AddToken        ();                 // Create the tokens for accepting states
BOOL    Clear           ();                 // Clear variables between two consecutive sub - DFA constructions
BOOL    AttachDFAs      ();                 // Build global DFA out of all sub - DFAs
void    CreateInitialPartition();           // Create initial partition for the DFA minimizing process
BOOL    MinimizeDFA     ();                 // Minimize the global DFA
BOOL    Node            (WORD c);           // Compute a node of the syntax tree
BOOL    ConstructDfa    ();                 // Construction of final DFA for the set of regular expressions
BOOL    ReadCode        (LPHUGE hpCode,
                         LPDWORD lpIndex);  // Read a fragment of code between { and }
BOOL    ReadId          (LPHUGE hpId,
                         LPDWORD lpIndex);  // Read an identifier
BOOL    ReadWhiteSpace  (LPHUGE hpWhiteSpace,
                         BOOL ReadEOL,
                         LPDWORD lpIndex);  // Read whitespace and comments
BOOL    ReadComment     (LPHUGE hpComment,
                         LPDWORD lpIndex);  // Read a comment
void    AddClassTransitions();              // Replace class transitions by character transitions
unsigned char BackslashConstant(LPHUGE s,
                            LPWORD Length); // Return the value of an escape constant beginning by backslash
BOOL    PackDFA         ();                 // Build packed arrays for global DFA
BOOL    WriteToFile     (HFILE hf,
                         LPSTR Buf);        // Write a string to a file

                                    /* Internal functions: dynamic arrays */

/* Pool management */
BOOL    PoolPush        (DWORD Position);   // Increment the top of the pool, and store a position on the new top
BOOL    PoolAddPosToList(DWORD List,
                         long int Position,
                         LPDWORD NewHead);  // Add an element to a linked list
void    PoolFreeList    (DWORD List);       // Free a linked list
DWORD   PoolListLength  (DWORD List);       // Length of a linked list

/* Others */
BOOL    PushStack       ();                 // The stack
BOOL    PushPosition    ();                 // The position array
BOOL    AddCharExp      (unsigned char c);  // The expanded regular expression array
BOOL    AddState        ();                 // The state table
BOOL    AddTransitionG  ();                 // The G transition table
BOOL    AddTransitionF  ();                 // The F transition table
BOOL    AddP2RegularExpression();           // The regular expression list of Part 2 of the source lex file
BOOL    AddP1Declaration();                 // The declarations list of Part 1 of the source lex file
BOOL    AddSymbol       ();                 // The table of grammar symbols to build the parser
BOOL    AddProduction   ();                 // The table of productions
BOOL    AddCsi          ();                 // The collection of sets of items
BOOL    AddActionGoto   ();                 // The action - goto table
BOOL    AllocateClass   ();                 // The character class table
BOOL    AddPropa        ();                 // The propagation table
